import sys
sys.path.append('..')

from neml import visco_flow, surfaces, hardening
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonGFlow(object):
  """
    Common to our Perzyna g functions.
  """
  def test_dg(self):
    s = self.gen_f()

    exact = self.model.dg(s, self.T)

    dfn = lambda x: self.model.g(x, self.T)
    num = differentiate(dfn, s)

    self.assertTrue(np.isclose(exact, num))

class TestGPowerLaw(unittest.TestCase, CommonGFlow):
  def setUp(self):
    self.n = 11.0
    self.eta = 200.0
    self.model = visco_flow.GPowerLaw(self.n, self.eta)
    self.T = 300.0

  def gen_f(self):
    return 100*0.25

  def test_properties(self):
    self.assertTrue(np.isclose(self.n, self.model.n(self.T)))
    self.assertTrue(np.isclose(self.eta, self.model.eta(self.T)))

  def test_g(self):
    s = self.gen_f()
    if (s > 0.0):
      v = (s/self.eta)**self.n
    else:
      v = 0.0
    self.assertTrue(np.isclose(self.model.g(s, self.T), v))

class CommonFlowRule(object):
  """
    Tests common to all flow rules.
  """
  def gen_stress(self):
    s = np.array([200,0,100,0,-50,0])
    return s

  def test_history(self):
    self.assertEqual(self.model.nhist, len(self.hist0))
    self.assertTrue(np.allclose(self.model.init_hist(), self.hist0))

  def test_dy_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.y(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dy_ds(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3, atol = 1.0e-6))

  def test_dy_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.y(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dy_da(stress, hist, self.T)
 
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.g(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dg_ds(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.g(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dg_da(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_ds_time(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.g_time(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dg_ds_time(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_da_time(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.g_time(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dg_da_time(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_ds_temp(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.g_temp(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dg_ds_temp(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dg_da_temp(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.g_temp(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dg_da_temp(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_ds(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.h(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dh_ds(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_da(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.h(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dh_da(stress, hist, self.T)

    #print(np.abs(exact))

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_ds_time(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.h_time(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dh_ds_time(stress, hist, self.T)
   
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_da_time(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.h_time(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dh_da_time(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_ds_temp(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda s: self.model.h_temp(s, hist, self.T)
    num = differentiate(dfn, stress)
    exact = self.model.dh_ds_temp(stress, hist, self.T)
    
    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))

  def test_dh_da_temp(self):
    stress = self.gen_stress()
    hist = self.gen_hist()

    dfn = lambda a: self.model.h_temp(stress, a, self.T)
    num = differentiate(dfn, hist)
    exact = self.model.dh_da_temp(stress, hist, self.T)

    self.assertTrue(np.allclose(num, exact, rtol = 1.0e-3))


class TestPerzynaIsoJ2Voce(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    self.s0 = 180.0
    self.R = 150.0
    self.d = 10.0

    self.n = 5.0
    self.eta = 20.0

    self.surface = surfaces.IsoJ2()
    self.hrule = hardening.VoceIsotropicHardeningRule(self.s0, self.R, self.d)
    self.g = visco_flow.GPowerLaw(self.n, self.eta)

    self.model = visco_flow.PerzynaFlowRule(self.surface, self.hrule, self.g)

    self.hist0 = np.zeros((1,))
    self.T = 300.0

  def gen_hist(self):
    return np.array([0.01])

class TestPerzynaIsoJ2Linear(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    self.s0 = 100.0
    self.Kp = 1500.0

    self.n = 2.0
    self.eta = 100.0

    self.surface = surfaces.IsoJ2()
    self.hrule = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    self.g = visco_flow.GPowerLaw(self.n, self.eta)

    self.model = visco_flow.PerzynaFlowRule(self.surface, self.hrule, self.g)

    self.hist0 = np.zeros((1,))
    self.T = 300.0

  def gen_hist(self):
    return np.array([0.01])

class TestChabocheModel(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    n = 20.0
    eta = 108.0
    sY = 89.0

    Q = 165.0
    b = 12.0
    
    self.m = 3

    C1 = 80.0e3
    C2 = 14.02e3
    C3 = 3.333e3

    y1 = 0.9e3
    y2 = 1.5e3
    y3 = 1.0

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(sY, Q, b)
    cs = [C1, C2, C3]
    gs = [y1, y2, y3]
    gmodels = [hardening.ConstantGamma(g) for g in gs]
    A = [0.0, 0.0, 0.0]
    ae = [1.0, 1.0, 1.0]

    hmodel = hardening.Chaboche(iso, cs, gmodels, A, ae)

    fluidity = visco_flow.ConstantFluidity(eta)

    self.hist0 = np.zeros((19,))
    self.T = 300.0

    self.model = visco_flow.ChabocheFlowRule(
        surface, hmodel, fluidity, n)

  def gen_hist(self):
    h = np.array([0.05,20,-30,40.0,5.0,2.0,40.0,-10,-20,10,50,30,10,-50,60,70,15,-15,10])
    h[1:7] = make_dev(h[1:7])
    h[7:13] = make_dev(h[7:13])
    h[13:19] = make_dev(h[13:19])
    return h

class TestChebocheFlow(unittest.TestCase):
  def setUp(self):
    self.n = 20.0
    self.eta = 108.0
    self.sY = 89.0

    self.Q = 165.0
    self.b = 12.0
    
    self.m = 3

    C1 = 80.0e3
    C2 = 14.02e3
    C3 = 3.333e3

    y1 = 0.9e3
    y2 = 1.5e3
    y3 = 1.0

    surface = surfaces.IsoKinJ2()
    self.iso = hardening.VoceIsotropicHardeningRule(self.sY, self.Q, self.b)
    self.cs = [C1, C2, C3]
    self.gs = [y1, y2, y3]
    self.As = [0.0, 0.0, 0.0]
    self.ns = [1.0, 1.0, 1.0]
    self.gmodels = [hardening.ConstantGamma(g) for g in self.gs]
    self.hmodel = hardening.Chaboche(self.iso, self.cs, self.gmodels, 
        self.As, self.ns)

    self.fluidity = visco_flow.ConstantFluidity(self.eta)

    self.hist0 = np.zeros((19,))
    self.T = 300.0

    self.model = visco_flow.ChabocheFlowRule(
        surface, self.hmodel, self.fluidity, self.n)

  def gen_stress(self):
    return np.array([300.0,-200.0,100.0,50.0,150.0,300.0])

  def gen_hist(self):
    h = np.array(range(1,self.m*6+2)) / (self.m*7)
    h[1:self.m*6+1] = (2.0 * h[1:self.m*6+1] - 1.0) * 200
    for i in range(self.m):
      h[1+i*6:1+(i+1)*6] = make_dev(h[1+i*6:1+(i+1)*6])
    return h

  def test_y(self):
    s = self.gen_stress()
    h = self.gen_hist()
    y = self.model.y(s, h, self.T)

    sdev = make_dev(s)
    X = sum(h[1+i*6:1+(i+1)*6] for i in range(self.m))
    ih = self.iso.q(h[0:1], self.T)[0]

    tv = np.sqrt(3.0/2.0) * la.norm(sdev + X) + ih

    if tv > 0.0:
      pd = (tv / self.eta)**self.n
    else:
      pd = 0.0

    gd = pd / np.sqrt(2.0/3.0)

    self.assertTrue(np.isclose(gd, y))

  def test_ep(self):
    s = self.gen_stress()
    h = self.gen_hist()
    y = self.model.y(s, h, self.T)

    ep_model = self.model.g(s, h, self.T) * y

    sdev = make_dev(s)
    X = sum(h[1+i*6:1+(i+1)*6] for i in range(self.m))
    ih = self.iso.q(h[0:1], self.T)[0]

    tv = np.sqrt(3.0/2.0) * la.norm(sdev + X) + ih

    if tv > 0.0:
      pd = (tv / self.eta)**self.n
    else:
      pd = 0.0

    ep_calc = 3.0/2.0 * pd * (sdev + X) / (np.sqrt(3.0/2.0) * la.norm(sdev+X))

    self.assertTrue(np.allclose(ep_calc, ep_model))

  def test_rdot(self):
    s = self.gen_stress()
    h = self.gen_hist()
    y = self.model.y(s, h, self.T)
    adot = self.model.h(s,h,self.T)[0] * y
    dr = -self.b * np.exp(-self.b * h[0]) * self.Q * adot

    sdev = make_dev(s)
    X = sum(h[1+i*6:1+(i+1)*6] for i in range(self.m))
    ih = self.iso.q(h[0:1], self.T)[0]

    tv = np.sqrt(3.0/2.0) * la.norm(sdev + X) + ih

    if tv > 0.0:
      pd = (tv / self.eta)**self.n
    else:
      pd = 0.0

    q = self.hmodel.q(h, self.T)

    Rtot = -self.sY - self.Q * (1.0 - np.exp(-self.b*h[0]))

    self.assertTrue(np.isclose(Rtot, q[0]))

    R_act = -(Rtot + self.sY)

    dr_direct = -self.b*(self.Q - R_act) * pd

    self.assertTrue(np.isclose(dr_direct, dr))

  def test_Xdot(self):
    s = self.gen_stress()
    h = self.gen_hist()
    y = self.model.y(s, h, self.T)
    dX = self.model.h(s, h, self.T)[1:] * y

    sdev = make_dev(s)
    X = sum(h[1+i*6:1+(i+1)*6] for i in range(self.m))
    ih = self.iso.q(h[0:1], self.T)[0]

    tv = np.sqrt(3.0/2.0) * la.norm(sdev + X) + ih

    if tv > 0.0:
      pd = (tv / self.eta)**self.n
    else:
      pd = 0.0

    ep_calc = 3.0/2.0 * pd * (sdev + X) / (np.sqrt(3.0/2.0) * la.norm(sdev+X))

    dX_calc = np.zeros((6*self.m,))

    for i in range(self.m):
      dX_calc[i*6:(i+1)*6] = -2.0/3.0 * self.cs[i] * ep_calc - self.gs[i] * h[1+i*6:1+(i+1)*6] * pd

    self.assertTrue(np.allclose(dX, dX_calc))


class CommonFluidity(object):
  """
    Common fluidity model tests
  """
  def gen_hist(self):
    return 0.15

  def test_deta(self):
    a = self.gen_hist()
    fm = self.model.deta(a, self.T)
    dfn = lambda a: self.model.eta(a, self.T)
    fn = differentiate(dfn, a)
    self.assertTrue(np.isclose(fm, fn))

class TestConstantFluidity(unittest.TestCase, CommonFluidity):
  def setUp(self):
    self.eta = 200.0
    self.model = visco_flow.ConstantFluidity(self.eta)
    self.T = 300.0

  def test_eta(self):
    a = self.gen_hist()
    self.assertTrue(np.isclose(self.eta, self.model.eta(a, self.T)))

class TestSaturatingFluidity(unittest.TestCase, CommonFluidity):
  def setUp(self):
    self.K0 = 116.0
    self.A = 238.9
    self.b = 9.80
    self.model = visco_flow.SaturatingFluidity(self.K0, self.A, self.b)
    self.T = 300.0

  def test_eta(self):
    a = self.gen_hist()
    self.assertTrue(np.isclose(self.K0 + self.A * (1.0 - np.exp(-self.b * a)), self.model.eta(a, self.T)))

class TestChabocheJ2Voce(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    self.n = 12.0
    self.K = 150.0
    self.k = 6.0
    self.C = 24800.0
    self.g0 = 300.0
    self.Q = 86 - self.k
    self.gs = 300.0
    self.b = 10.0
    self.beta = 0.0

    self.surface = surfaces.IsoKinJ2()

    self.iso = hardening.VoceIsotropicHardeningRule(self.k, self.Q, self.b)
    cs = [self.C]
    gs = [hardening.SatGamma(self.gs, self.g0, self.beta)]
    As = [0.0]
    ns = [1.0]
    self.hardening = hardening.Chaboche(self.iso, cs, gs, As, ns)

    self.fluidity = visco_flow.ConstantFluidity(self.K)

    self.model = visco_flow.ChabocheFlowRule(
        self.surface, self.hardening, self.fluidity, self.n) 

    self.hist0 = np.zeros((7,))
    self.T = 300.0

  def gen_hist(self):
    bs = make_dev([10.0, -50.0, 25.0, 30.0, -10.0, 30.0])
    return np.array([0.01] + list(bs))

class TestYaguchiGr91Flow(unittest.TestCase, CommonFlowRule):
  def setUp(self):
    self.model = visco_flow.YaguchiGr91FlowRule()

    self.pTrange = np.linspace(473.0, 873.0)

    self.hist0 = np.zeros((14,))
    self.T = 500.0

  def gen_stress(self):
    s = np.array([200,75,100,50,-50,100])*0.1
    return s
  
  @staticmethod
  def J2(x):
    return np.sqrt(3.0/2.0 * np.dot(x,x))

  def test_y(self):
    s = self.gen_stress()
    h = self.gen_hist()

    y_model = self.model.y(s, h, self.T)

    X = h[:6] + h[6:12]
    Xdev = make_dev(X)
    Sdev = make_dev(s)

    t1 = (self.J2(Xdev - Sdev) - h[13])/self.model.D(self.T)
    if t1 > 0.0:
      y_calc = t1**self.model.n(self.T)
    else:
      y_calc = 0.0

    self.assertTrue(np.isclose(y_calc, y_model))

  def test_g(self):
    s = self.gen_stress()
    h = self.gen_hist()

    g_model = self.model.g(s, h, self.T)

    X = h[:6] + h[6:12]
    Xdev = make_dev(X)
    Sdev = make_dev(s)

    g_calc = 3.0/2.0 * (Sdev - Xdev) / self.J2(Sdev - Xdev)

    self.assertTrue(np.allclose(g_calc, g_model))

  def test_h(self):
    s = self.gen_stress()
    a = self.gen_hist()

    h_model = self.model.h(s, a, self.T)
    h_calc = np.zeros(h_model.shape)
    
    C1 = self.model.C1(self.T)
    C2 = self.model.C2(self.T)

    a1 = self.model.a10(self.T) - a[12]
    a2 = self.model.a2(self.T)

    X = a[:6] + a[6:12]
    Xdev = make_dev(X)
    Sdev = make_dev(s)

    n = 3.0/2.0 * (Sdev - Xdev) / self.J2(Sdev - Xdev)

    h_calc[:6] = C1 * (2.0/3.0 * a1 * n - a[:6])

    h_calc[6:12] = C2 * (2.0/3.0 * a2 * n - a[6:12])

    h_calc[12] = self.model.d(self.T) * (self.model.q(self.T) - a[12])

    t1 = (self.J2(Xdev - Sdev) - a[13])/self.model.D(self.T)
    if t1 > 0.0:
      y_calc = t1**self.model.n(self.T)
    else:
      y_calc = 0.0

    sas = self.model.A(self.T) + self.model.B(self.T) * np.log10(y_calc)
    if sas < 0.0:
      sas = 0.0

    if (sas - a[13]) >= 0:
      b = self.model.bh(self.T)
    else:
      b = self.model.br(self.T)

    h_calc[13] = b * (sas - a[13])

    self.assertTrue(np.allclose(h_calc, h_model))

  def gen_hist(self):
    X1 = make_dev([15.0,-20.0,-30.0,50.0,-10.0,5])
    X2 = make_dev([-25,10,15,30,-15,20])
    return np.array(list(X1) + list(X2) + [50.0] + [5.0]) / 10.0

  def eval_prop(self, prop):
    return np.array([prop(T) for T in self.pTrange])

  def test_D(self):
    Ds_model = self.eval_prop(self.model.D)
    Ds_direct = (-1.119e-9 + 5.145e-11 * self.pTrange - 5.450e-14 * self.pTrange**2.0)**(-1.0/2.85)
    self.assertTrue(np.allclose(Ds_model, Ds_direct))
  
  def test_n(self):
    ns_model = self.eval_prop(self.model.n)
    ns_direct = 2.850 * np.ones(self.pTrange.shape)
    self.assertTrue(np.allclose(ns_model, ns_direct))

  def test_a10(self):
    a10s_model = self.eval_prop(self.model.a10)
    a10s_direct = 2.082e3 - 8.110*self.pTrange + 1.321e-2*self.pTrange**2.0 - 7.278e-6 * self.pTrange**3.0
    self.assertTrue(np.allclose(a10s_model, a10s_direct))

  def test_C2(self):
    C2s_model = self.eval_prop(self.model.C2)
    C2s_direct = np.piecewise(self.pTrange, [self.pTrange < 798.0, self.pTrange >= 798.0], [lambda x: 2.000e2, lambda x: -2.992e3 + 4.0 * x])
    self.assertTrue(np.allclose(C2s_model, C2s_direct))

  def test_a2(self):
    a2s_model = self.eval_prop(self.model.a2)
    a2s_direct = np.piecewise(self.pTrange, [self.pTrange < 773.0, self.pTrange >= 773.0], [lambda x: 1.100e2, lambda x: 3.373e3 - 7.622*x + 4.400e-3*x**2.0])
    self.assertTrue(np.allclose(a2s_model, a2s_direct))

  def test_g1(self):
    g1s_model = self.eval_prop(self.model.g1)
    g1s_direct = np.piecewise(self.pTrange, [self.pTrange < 773.0, self.pTrange >= 773.0], [lambda x: 5.064e-137 * np.exp(3.545e-1 * x), lambda x: 5.336e-33 * np.exp(4.470e-2 * x)])
    self.assertTrue(np.allclose(g1s_model, g1s_direct))

  def test_g2(self):
    g2s_model = self.eval_prop(self.model.g2)
    g2s_direct = np.piecewise(self.pTrange, [self.pTrange < 773.0, self.pTrange >= 773.0], [lambda x: 8.572e-108 * np.exp(2.771e-1*x), lambda x: 2.817e-11 - 7.538e-14 * x + 5.039e-17 * x**2.0])
    self.assertTrue(np.allclose(g2s_model, g2s_direct))
  
  def test_m(self):
    ms_model = self.eval_prop(self.model.m)
    ms_direct = np.piecewise(self.pTrange, [self.pTrange < 673.0, np.logical_and(673.0 <= self.pTrange, self.pTrange < 773.0), self.pTrange >= 773.0], [lambda x: 1.200e1, lambda x: 5.766e2*np.exp(-5.754e-3*x), lambda x: 6.750])
    self.assertTrue(np.allclose(ms_model, ms_direct))
  
  def test_br(self):
    brs_model = self.eval_prop(self.model.br)
    brs_direct = 1.000e3 * np.ones(self.pTrange.shape)
    self.assertTrue(np.allclose(brs_model, brs_direct))

  def test_bh(self):
    bhs_model = self.eval_prop(self.model.bh)
    bhs_direct = 1.065e3 - 1.404 * self.pTrange + 4.417e-4 * self.pTrange**2.0
    self.assertTrue(np.allclose(bhs_model, bhs_direct))

  def test_A(self):
    As_model = self.eval_prop(self.model.A)
    As_direct = np.piecewise(self.pTrange, [self.pTrange < 673.0, np.logical_and(673.0 <= self.pTrange, self.pTrange < 823.0), self.pTrange >= 823.0], [lambda x: -1.841e2 + 2.075e-1 * x, lambda x: -4.799e2 + 1.262*x - 9.133e-4 * x**2.0, lambda x: -6.0e1])
    self.assertTrue(np.allclose(As_model, As_direct))

  def test_B(self):
    Bs_model = self.eval_prop(self.model.B)
    Bs_direct = np.piecewise(self.pTrange, [self.pTrange < 773.0, self.pTrange >= 773.0], [lambda x: -6.340e1 + 6.850e-2*x, lambda x: -1.730e1])
    self.assertTrue(np.allclose(Bs_model, Bs_direct))

  def test_d(self):
    ds_model = self.eval_prop(self.model.d)
    ds_direct = 3.208 - 1.010e-2 * self.pTrange + 1.012e-5 * self.pTrange**2.0
    self.assertTrue(np.allclose(ds_model, ds_direct))

  def test_q(self):
    qs_model = self.eval_prop(self.model.q)
    qs_direct = np.piecewise(self.pTrange, [self.pTrange < 673.0, np.logical_and(673.0 <= self.pTrange, self.pTrange < 823.0), self.pTrange >= 823.0], [lambda x: 9.50e1, lambda x: -2.467e2 + 7.320e-1*x - 3.333e-4*x**2.0, lambda x: 1.300e2])
    self.assertTrue(np.allclose(qs_model, qs_direct))

  def test_C1(self):
    C1s_model = self.eval_prop(self.model.C1)
    C1s_direct = np.piecewise(self.pTrange, [self.pTrange < 673.0, np.logical_and(673.0 <= self.pTrange, self.pTrange < 773.0), self.pTrange >= 773.0], [lambda x: 1.50e3, lambda x: -2.879e4 + 45.0*x, lambda x: 6.000e3])
    self.assertTrue(np.allclose(C1s_model, C1s_direct))




from __future__ import division

import numpy as np

from common import *

from neml.math import tensors
from neml import history

def diff_skew_symmetric(fn, s0):
  dfn = lambda s: fn(tensors.Symmetric(usym(s))).data

  return tensors.SkewSymR4(differentiate(dfn, s0.data))

def diff_skew_history(fn, h0):
  vec = np.copy(np.array(h0))

  def dfn(x):
    H = h0.deepcopy()
    H.copy_data(x)
    return fn(H).data

  return differentiate(dfn, vec)

def diff_symmetric_symmetric(fn, s0):
  dfn = lambda s: fn(tensors.Symmetric(usym(s))).data

  return tensors.SymSymR4(differentiate(dfn, s0.data))

def diff_symmetric_skew(fn, w0):
  dfn = lambda w: fn(tensors.Skew(uskew(w))).data

  return tensors.SymSkewR4(differentiate(dfn, w0.data))

def diff_symmetric_history(fn, h0):
  vec = np.copy(np.array(h0))

  def dfn(x):
    H = h0.deepcopy()
    H.copy_data(x)
    return fn(H).data

  nd = differentiate(dfn, vec)

  return nd

def diff_scalar_symmetric(fn, S0):
  dfn = lambda s: fn(tensors.Symmetric(s))
  
  return tensors.Symmetric(differentiate(dfn, usym(S0.data))[0])

def diff_history_scalar(fn, h0):
  vec = np.copy(np.array(h0))
  
  def dfn(x):
    H = h0.deepcopy()
    H.copy_data(x)
    return fn(H)

  nd = differentiate(dfn, vec)

  H = h0.deepcopy()
  H.copy_data(nd)

  return H

def diff_history_symmetric(fn, s0):
  dfn = lambda s: np.array(fn(tensors.Symmetric(usym(s))))
  
  res = differentiate(dfn, s0.data)

  return res

def diff_history_skew(fn, w0):
  dfn = lambda w: np.array(fn(tensors.Skew(uskew(w))))

  return differentiate(dfn, w0.data)

def diff_history_history(fn, h0):
  vec = np.copy(np.array(h0))

  def dfn(x):
    H = h0.deepcopy()
    H.copy_data(x)
    return np.array(fn(H))

  nd = differentiate(dfn, vec)

  return nd



import sys
sys.path.append('..')

from neml import larsonmiller, interpolate
import unittest

from common import *

import numpy as np

class TestLarsonMiller(unittest.TestCase):
  def setUp(self):
    self.fn = interpolate.PolynomialInterpolate([-6.653e-9,2.952e-4,-6.197e-1])
    self.C = 32.06

    self.lmr = larsonmiller.LarsonMillerRelation(self.fn, self.C)

    self.T = 550 + 273.15

  def test_sR(self):
    tR = 50000.0
    lmp = self.T*(self.C + np.log10(tR))
    direct = 10.0**(self.fn.value(lmp))
    test = self.lmr.sR(tR, self.T)
    self.assertTrue(np.isclose(direct, test))

  def test_tR(self):
    tR = 50000.0
    lmp = self.T*(self.C + np.log10(tR))
    sR = 10.0**(self.fn.value(lmp))
    tR2 = self.lmr.tR(sR, self.T)
    self.assertTrue(np.isclose(tR2, tR))

  def test_dtR(self):
    sR = 200.0
    num = differentiate(lambda s: self.lmr.tR(s, self.T), sR)
    act = self.lmr.dtR_ds(sR, self.T)
    self.assertTrue(np.isclose(num, act, rtol = 1e-4))



#!/usr/bin/env python3

from neml import history, interpolate
from neml.math import tensors, rotations, matrix
from neml.cp import crystallography, slipharden, sliprules

from common import differentiate
from nicediff import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonSlipHardening():
  def test_d_hist_d_stress(self):
    d = np.array(self.model.d_hist_d_s(self.S, self.Q, self.H, self.L, self.T, self.sliprule, self.fixed))
    nd = diff_history_symmetric(lambda s: self.model.hist(s, self.Q, self.H, self.L, self.T,
      self.sliprule, self.fixed), self.S)
    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

  def test_d_hist_d_hist(self):
    d = np.array(self.model.d_hist_d_h(self.S, self.Q, self.H, self.L, self.T, self.sliprule, 
      self.fixed))
    nd = diff_history_history(lambda h: self.model.hist(self.S, self.Q, h, self.L, self.T,
      self.sliprule, self.fixed), self.H)

    print(d)
    print(nd)

    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

  def test_d_hist_to_tau_d_hist(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        nd = diff_history_scalar(lambda h: self.model.hist_to_tau(g, i, h, self.L, self.T, self.fixed), self.H)
        d = self.model.d_hist_to_tau(g, i, self.H, self.L, self.T, self.fixed)
        self.assertTrue(np.allclose(np.array(nd), np.array(d)))

class TestConstantHardening(unittest.TestCase, CommonSlipHardening):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    
    self.nslip = self.L.ntotal

    self.static = 20.0

    self.H = history.History()

    self.T = 300.0

    self.s0 = [self.static]*self.nslip

    self.model = slipharden.FixedStrengthHardening(self.s0)

    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

  def test_hist_to_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        model = self.model.hist_to_tau(g, i, self.H, self.L, self.T,
            self.fixed)
        should = self.static
        self.assertAlmostEqual(model, should)

  def test_definition(self):
    hrate = self.model.hist(self.S, self.Q, self.H, self.L, self.T, self.sliprule,
        self.fixed)
    exact = np.array([])
    self.assertTrue(np.allclose(hrate, exact))

class TestGeneralLinearHardeningNoAbs(unittest.TestCase, CommonSlipHardening):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    
    self.nslip = self.L.ntotal

    self.static = 20.0
    self.current = 25.0

    self.H = history.History()
    for i in range(self.nslip):
      self.H.add_scalar("strength"+str(i))
      self.H.set_scalar("strength"+str(i), self.current)

    self.T = 300.0

    self.M = matrix.SquareMatrix(self.nslip, type = "block", 
        data = [0.1,0.2,0.3,0.4], blocks = [6,6])

    self.s0 = [self.static]*self.nslip

    self.model = slipharden.GeneralLinearHardening(self.M, self.s0, 
        absval = False)

    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

  def test_hist_to_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        model = self.model.hist_to_tau(g, i, self.H, self.L, self.T,
            self.fixed)
        should = self.static + self.current
        self.assertAlmostEqual(model, should)

  def test_definition(self):
    hrate = self.model.hist(self.S, self.Q, self.H, self.L, self.T, self.sliprule,
        self.fixed)
    srates = [self.sliprule.slip(g, i, self.S, self.Q, self.H, self.L, self.T, 
      self.fixed) for g in range(self.L.ngroup) for i in range(self.L.nslip(g))]
    exact = np.dot(np.array(self.M), srates)
    self.assertTrue(np.allclose(hrate, exact))

class TestGeneralLinearHardeningAbs(unittest.TestCase, CommonSlipHardening):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    
    self.nslip = self.L.ntotal

    self.static = 20.0
    self.current = 25.0

    self.H = history.History()
    for i in range(self.nslip):
      self.H.add_scalar("strength"+str(i))
      self.H.set_scalar("strength"+str(i), self.current)

    self.T = 300.0

    self.M = matrix.SquareMatrix(self.nslip, type = "block", 
        data = [0.1,0.2,0.3,0.4], blocks = [6,6])

    self.s0 = [self.static]*self.nslip

    self.model = slipharden.GeneralLinearHardening(self.M, self.s0, 
        absval = True)

    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

  def test_hist_to_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        model = self.model.hist_to_tau(g, i, self.H, self.L, self.T,
            self.fixed)
        should = self.static + self.current
        self.assertAlmostEqual(model, should)

  def test_definition(self):
    hrate = self.model.hist(self.S, self.Q, self.H, self.L, self.T, self.sliprule,
        self.fixed)
    srates = [np.abs(self.sliprule.slip(g, i, self.S, self.Q, self.H, self.L, self.T, 
      self.fixed)) for g in range(self.L.ngroup) for i in range(self.L.nslip(g))]
    exact = np.dot(np.array(self.M), srates)
    self.assertTrue(np.allclose(hrate, exact))

class CommonSlipSingleHardening():
  def test_d_hist_map(self):
    nd = diff_history_scalar(lambda h: self.model.hist_map(h, self.T, 
      self.fixed), self.H)
    H = self.model.d_hist_map(self.H, self.T, self.fixed)
    self.assertTrue(np.allclose(np.array(H), np.array(nd)))

  def test_hist_to_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        self.assertTrue(np.isclose(self.strength + self.static + self.nye_part, 
          self.model.hist_to_tau(g, i, self.H, self.L, self.T, self.fixed)))

class CommonSlipSingleStrengthHardening():
  def test_hist_map(self):
    self.assertTrue(np.isclose(self.model.hist_map(self.H, self.T,
      self.fixed), self.strength + self.static + self.nye_part))

  def test_hist(self):
    h = self.model.hist(self.S, self.Q, self.H, self.L, self.T, self.sliprule, self.fixed)
    self.assertTrue(np.isclose(h.get_scalar(self.vname), self.model.hist_rate(
      self.S, self.Q, self.H, self.L, self.T, self.sliprule, self.fixed)))

  def test_d_hist_rate_d_stress(self):
    dfn = lambda s: self.model.hist_rate(s, self.Q, self.H,
        self.L, self.T, self.sliprule, self.fixed)
    nd = diff_scalar_symmetric(dfn, self.S)
    d = self.model.d_hist_rate_d_stress(self.S, self.Q, self.H, self.L,
        self.T, self.sliprule, self.fixed)
    self.assertEqual(d,nd)

  def test_d_hist_rate_d_hist(self):
    dfn = lambda h: self.model.hist_rate(self.S, self.Q, h, 
        self.L, self.T, self.sliprule, self.fixed)
    nd = diff_history_scalar(dfn, self.H)
    
    d = self.model.d_hist_rate_d_hist(self.S, self.Q, self.H, self.L,
        self.T, self.sliprule, self.fixed)
    
    self.assertTrue(np.isclose(d.get_scalar(self.vname), 
      nd.get_scalar(self.vname)))

  def test_nye_parts_match(self):
    self.assertTrue(np.isclose(self.model.nye_part(self.nye, self.T), 
      self.model.nye_contribution(self.fixed, self.T)))

  def test_no_nye_works(self):
    self.assertTrue(np.isclose(self.model.nye_contribution(history.History(), self.T),
      0.0))

class CommonPlasticSlipHardening():
  def sum_slip(self):
    return sum(np.abs(self.sliprule.slip(g, i, self.S, self.Q, self.H, self.L, 
      self.T, self.fixed)) for g in range(self.L.ngroup) for i in range(self.L.nslip(g)))

  def test_hist_rate(self):
    ss = self.sum_slip()

    self.assertTrue(np.isclose(self.model.hist_rate(self.S, self.Q, self.H,
      self.L, self.T, self.sliprule, self.fixed), self.model.hist_factor(self.H.get_scalar(self.vname), 
        self.L, self.T, self.fixed) * ss))

  def test_d_factor(self):
    nd = differentiate(lambda s: self.model.hist_factor(s, self.L, self.T, self.fixed), 
        self.strength)
    d = self.model.d_hist_factor(self.strength, self.L, self.T, self.fixed)

    self.assertTrue(np.isclose(nd, d))
    
class TestVoceHardening(unittest.TestCase, CommonPlasticSlipHardening, 
    CommonSlipSingleStrengthHardening, CommonSlipSingleHardening, CommonSlipHardening):
  def setUp(self):
    self.vname = "strength"

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.T = 300.0

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.static = self.tau0

    self.model = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

    self.nye = tensors.RankTwo([[1.1,1.2,1.3],[2.1,2.2,2.3],[3.1,3.2,3.3]])

    self.nye_part = 0.0

  def test_initialize_hist(self):
    H = history.History()
    self.model.populate_history(H)
    self.model.init_history(H)
    self.assertTrue(np.isclose(H.get_scalar('strength'), 0.0))

  def test_static_strength(self):
    self.assertTrue(np.isclose(self.model.static_strength(self.T), self.tau0))
  
  def test_factor(self):
    self.assertTrue(np.isclose(
      self.model.hist_factor(self.strength, self.L, self.T, self.fixed),
      self.b * (self.tau_sat - self.strength)))

  def test_nye_factor(self):
    self.assertTrue(np.isclose(
      self.model.nye_part(self.nye, self.T), 0.0))

  def test_use_nye(self):
    self.assertFalse(self.model.use_nye)

class TestNyeVoceHardening(unittest.TestCase, CommonPlasticSlipHardening, 
    CommonSlipSingleStrengthHardening, CommonSlipSingleHardening, CommonSlipHardening):
  def setUp(self):
    self.vname = "strength"

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.T = 300.0

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.static = self.tau0

    self.k = 1.2

    self.model = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0, 
        k = self.k)
    
    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.nye = tensors.RankTwo([[1.1,1.2,1.3],[2.1,2.2,2.3],[3.1,3.2,3.3]])
    self.nye_part = self.k * np.sqrt(la.norm(self.nye.data.reshape((3,3)), ord = 'fro'))

    self.fixed = history.History()
    self.fixed.add_ranktwo("nye")
    self.fixed.set_ranktwo("nye", self.nye)
    
  def test_initialize_hist(self):
    H = history.History()
    self.model.populate_history(H)
    self.model.init_history(H)
    self.assertTrue(np.isclose(H.get_scalar('strength'), 0.0))

  def test_static_strength(self):
    self.assertTrue(np.isclose(self.model.static_strength(self.T), self.tau0))
  
  def test_factor(self):
    self.assertTrue(np.isclose(
      self.model.hist_factor(self.strength, self.L, self.T, self.fixed),
      self.b * (self.tau_sat - self.strength)))

class TestVoceHardeningMoveVar(unittest.TestCase, CommonPlasticSlipHardening, 
    CommonSlipSingleStrengthHardening, CommonSlipSingleHardening, CommonSlipHardening):
  def setUp(self):
    self.vname = "wee"
    
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar(self.vname)
    self.H.set_scalar(self.vname, self.strength)

    self.T = 300.0

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.static = self.tau0

    self.model = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.model.set_varnames([self.vname])
    
    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

    self.nye = tensors.RankTwo([[1.1,1.2,1.3],[2.1,2.2,2.3],[3.1,3.2,3.3]])

    self.nye_part = 0.0

  def test_get_vnames(self):
    self.assertEqual(self.model.varnames, [self.vname])

  def test_initialize_hist(self):
    H = history.History()
    self.model.populate_history(H)
    self.model.init_history(H)
    self.assertTrue(np.isclose(H.get_scalar(self.vname), 0.0))

  def test_static_strength(self):
    self.assertTrue(np.isclose(self.model.static_strength(self.T), self.tau0))
  
  def test_factor(self):
    self.assertTrue(np.isclose(
      self.model.hist_factor(self.strength, self.L, self.T, self.fixed),
      self.b * (self.tau_sat - self.strength)))

  def test_nye_factor(self):
    self.assertTrue(np.isclose(
      self.model.nye_part(self.nye, self.T), self.nye_part))

  def test_use_nye(self):
    self.assertFalse(self.model.use_nye)

class TestMultiVoceHardening(unittest.TestCase, 
    CommonSlipSingleHardening, CommonSlipHardening):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength_0 = 35.0
    self.strength_1 = -5.0
    self.H = history.History()
    self.H.add_scalar("strength0")
    self.H.set_scalar("strength0", self.strength_0)
    self.H.add_scalar("strength1")
    self.H.set_scalar("strength1", self.strength_1)

    self.T = 300.0

    self.tau0_1 = 10.0
    self.tau_sat_1 = 50.0
    self.b_1 = 2.5

    self.tau0_2 = 5.0
    self.tau_sat_2 = -25.0
    self.b_2 = 1.5

    self.static = self.tau0_1 + self.tau0_2
    self.strength = self.strength_0 + self.strength_1

    self.model1 = slipharden.VoceSlipHardening(self.tau_sat_1, self.b_1, self.tau0_1)
    self.model2 = slipharden.VoceSlipHardening(self.tau_sat_2, self.b_2, self.tau0_2)

    self.model = slipharden.SumSlipSingleStrengthHardening([self.model1,
      self.model2])
    
    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.fixed = history.History()

    self.nye = tensors.RankTwo([[1.1,1.2,1.3],[2.1,2.2,2.3],[3.1,3.2,3.3]])

    self.nye_part = 0.0

  def test_initialize_hist(self):
    H = history.History()
    self.model.populate_history(H)
    self.model.init_history(H)
    self.assertTrue(np.isclose(H.get_scalar('strength0'), 0.0))
    self.assertTrue(np.isclose(H.get_scalar('strength1'), 0.0))

  def test_hist_def(self):
    hrate_1 = self.b_1 * (self.tau_sat_1 - self.strength_0) * self.sliprule.sum_slip(self.S, self.Q, 
        self.H, self.L, self.T, self.fixed)
    hrate_2 = self.b_2 * (self.tau_sat_2 - self.strength_1) * self.sliprule.sum_slip(self.S, self.Q,
        self.H, self.L, self.T, self.fixed)

    hrates = np.array([hrate_1, hrate_2])

    hist = self.model.hist(self.S, self.Q, self.H, self.L, self.T, self.sliprule, self.fixed)

    self.assertTrue(np.allclose(hrates, hist))

  def test_use_nye(self):
    self.assertFalse(self.model.use_nye)

class TestLinearSlipHardening(unittest.TestCase, CommonPlasticSlipHardening, 
    CommonSlipSingleStrengthHardening, CommonSlipSingleHardening, CommonSlipHardening):
  def setUp(self):
    self.vname = "strength"

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.T = 300.0

    self.tau0 = 10.0
    self.k1 = 2.0
    self.k2 = 3.0

    self.static = self.tau0

    self.model = slipharden.LinearSlipHardening(self.tau0, self.k1, self.k2)
    
    self.g0 = 1.0
    self.n = 3.0
    self.sliprule = sliprules.PowerLawSlipRule(self.model, self.g0, self.n)

    self.nye = tensors.RankTwo([[1.1,1.2,1.3],[2.1,2.2,2.3],[3.1,3.2,3.3]])
    self.nye_part = self.k2 * la.norm(self.nye.data.reshape((3,3)), ord = 'fro')

    self.fixed = history.History()
    self.fixed.add_ranktwo("nye")
    self.fixed.set_ranktwo("nye", self.nye)
    
  def test_initialize_hist(self):
    H = history.History()
    self.model.populate_history(H)
    self.model.init_history(H)
    self.assertTrue(np.isclose(H.get_scalar('strength'), 0.0))

  def test_static_strength(self):
    self.assertTrue(np.isclose(self.model.static_strength(self.T), self.tau0))
  
  def test_factor(self):
    self.assertTrue(np.isclose(
      self.model.hist_factor(self.strength, self.L, self.T, self.fixed),
      self.k1))

  def test_nye_factor(self):
    self.assertTrue(np.isclose(
      self.model.nye_part(self.nye, self.T), self.nye_part))

  def test_use_nye(self):
    self.assertTrue(self.model.use_nye)




import numpy as np
import numpy.linalg as la
import networkx as nx

import warnings

import scipy.optimize as opt

from neml import uniaxial

class BarModel(nx.MultiGraph):
  """
    Model of an arbitrary bar network.

    Graph structure:
      I am an networkx graph

      Bars are edge data
      Nodes are rigid links

      DoFs are nodal displacements
      Equations are nodal equilibrium

      The specific input structure is a networkx graph.
      You may label the nodes however you want.
      Each edge must be a bar object, as defined below.

    Boundary conditions:
      Boundary conditions are at nodes
      They can be imposed forces or imposed displacements
      They are specified as functions of time

      Additionally, temperature histories can be assigned to 
      each bar using the bar object

    Loading:
      Calling solve(dt) advances the model from t_n to t_n + dt

    Solution history:
      Each bar maintains its own history internally, 
        as defined in the bar class
      Each node has properties:
          "displacements"
          "forces"

        that are maintained as the structure evolves

        The user does not need to provide these in the networkx model --
        they will be initialized in this object's __init__
      Additionally, the global properties:
          "time"
          "energy"
          "dissipation"
        are tracked 
  """
  def __init__(self, *args, **kwargs):
    super(nx.MultiGraph, self).__init__(*args, **kwargs)
    self.time = np.array([0.0])
    self.energy = np.array([0.0])
    self.dissipation = np.array([0.0])
    self.validated = False

    # Sign convention
    self.sign = [-1.0,1.0]

  def solve(self, dt, ndiv = 4, dfact = 2, verbose = False,
      extrapolate = False, rtol = 1.0e-6, atol = 1.0e-8):
    """
      Solve the next time increment

      Parameters:
        dt      time increment

      Optional:
        ndiv    max number of adaptive subdivisions
        dfact   what to divided by (SHOULD BE AN INTEGER)
        verbose dump a lot of solver info
    """
    if not self.validated:
      self.validate()
    
    free_nodes, fixed_nodes = self.free_fixed_nodes()

    dt_total = dt
    dt_attempt = dt
    step_total = dfact**ndiv
    step_attempt = dfact**ndiv
    step_curr = 0
    dtried = 0
    ef = 1.0

    while step_curr < step_total:
      solveme = lambda d: self.RJ(d, free_nodes, dt_attempt)
      d_guess = self.make_guess(free_nodes, extrapolate = extrapolate,
          value = ef)   
      try:
        d = newton(solveme, d_guess, verbose = verbose, fail_iter = True, 
            rtol = rtol, atol = atol)
        step_curr += step_attempt
      except Exception as e:
        dtried += 1
        if verbose:
          print("Substepping: %i" % dtried)
        if dtried == ndiv:
          raise e
        step_attempt /= dfact
        dt_attempt /= dfact
        ef /= dfact
        continue
      self.apply_update(d, free_nodes, dt_attempt)
      ef = 1.0

  def apply_update(self, d, free_nodes, dt):
    """
      Actually apply the results of an update

      Parameters:
        d           free displacements
        free_nodes  the list of free nodes
        dt          time increment
    """
    t_next = self.time[-1] + dt
    
    for n in self.nodes():
      # This gets summed below
      self.nodes[n]['forces'] = np.append(self.nodes[n]['forces'], 0.0)
      if n in free_nodes:
        self.nodes[n]['displacements'] = np.append(
            self.nodes[n]['displacements'], d[free_nodes.index(n)])
      else:
        self.nodes[n]['displacements'] = np.append(
            self.nodes[n]['displacements'], 
            self.nodes[n]['displacement bc'](t_next))
    
    e_sum = 0.0
    p_sum = 0.0
    for i,j,data in self.edges(data=True):
      me = data['object']
      for ni,n in enumerate((i,j)):
        self.nodes[n]['forces'][-1] += self.sign[ni] * me.stress_next * me.A
      e_sum += me.energy_next * me.A * me.l
      p_sum += me.dissipation_next * me.A * me.l
      me.advance_step()

    self.time = np.append(self.time, t_next)
    self.energy = np.append(self.energy, e_sum)
    self.dissipation = np.append(self.dissipation, e_sum)
  
  def make_guess(self, free_nodes, extrapolate = True, value = 1.0):
    """
      Make a displacement guess

      Parameters:
        free_nodes  the list of free nodes (this sets dof numbering)
    """
    if (not extrapolate) or (len(self.time) < 2):
      return np.array([self.nodes[n]['displacements'][-1] for n in free_nodes])
    else:
      xprev = np.array([self.nodes[n]['displacements'][-1] for n in free_nodes])
      xprevprev = np.array([self.nodes[n]['displacements'][-2] for n in free_nodes])
      return xprev + value * (xprev - xprevprev)
  
  def free_fixed_nodes(self):
    """
      Return lists of the free and fixed nodes
    """
    free = [n for n in self.nodes() if 'displacement bc' not in self.nodes[n]]
    fixed = [n for n in self.nodes() if 'displacement bc' in self.nodes[n]]

    return free, fixed

  def RJ(self, d, free_nodes, dt):
    """
      Compute the residual equilibrium equation and the Jacobian

      Parameters:
        d           current free displacements
        free_nodes  listing of free nodes (sets the dof order)
        dt          time increment
    """
    R = np.zeros((len(free_nodes),))
    J = np.zeros((len(free_nodes), len(free_nodes)))

    t_next = self.time[-1] + dt

    # External forces
    for i,n in enumerate(free_nodes):
      if 'force bc' in self.nodes[n]:
        R[i] -= self.nodes[n]['force bc'](t_next)
      
    for i,j,data in self.edges(data=True):
      # Delta d
      dd = 0.0
      for index, node in enumerate((i,j)):
        if node in free_nodes:
          dd += self.sign[index]*d[free_nodes.index(node)]
        else:
          dd += self.sign[index]*self.nodes[node]['displacement bc'](t_next)
      
      # Force update
      f, A = data['object'].update_force(dd, t_next, self.time[-1])

      # Assemble internal forces and derivative
      for index_i, node_i in enumerate((i,j)):
        if node_i in free_nodes:
          idx_i = free_nodes.index(node_i)
          R[idx_i] += self.sign[index_i] * f

          for index_j, node_j in enumerate((i,j)):
            if node_j in free_nodes:
              idx_j = free_nodes.index(node_j)
              J[idx_i, idx_j] += (self.sign[index_i]*self.sign[index_j] * A)
     
    return R, J

  def add_force_bc(self, node, bfn):
    """
      Add a force boundary condition

      Parameters:
        node        node to add to
        bfn         force as a function of time
    """
    if 'force bc' in self.nodes[node] or 'displacement bc' in self.nodes[node]:
      warnings.warn("Overriding previous boundary condition")
    if 'displacement bc' in self.nodes[node]:
      del self.nodes[node]['displacement bc']

    self.nodes[node]['force bc'] = bfn

  def add_displacement_bc(self, node, bfn):
    """
      Parameters:
        node        node to add to
        bfn         displacement as a function of time
    """
    if 'force bc' in self.nodes[node] or 'displacement bc' in self.nodes[node]:
      warnings.warn("Overriding previous boundary condition")
    if 'force bc' in self.nodes[node]:
      del self.nodes[node]['force bc']

    self.nodes[node]['displacement bc'] = bfn

  def validate(self):
    """
      Iterate through the graph and make sure that everything is okay to
      run.
    """
    for node in self.nodes():
      if 'displacements' not in self.nodes[node]:
        self.nodes[node]['displacements'] = np.array([0.0])
      if 'forces' not in self.nodes[node]:
        self.nodes[node]['forces'] = np.array([0.0])
    self.validated = True

  def gather_element(self, quantity):
    """
      Iterate through elements and gather a particular quantity

      Parameters:
        quantity        what to grab
    """
    return np.array([getattr(e['object'], quantity) for i,j,e in 
      self.edges(data = True)])

class Bar(object):
  """
    Object representing a single bar, including its history
  """
  def __init__(self, mat, A, l, T = lambda t: 0.0):
    """
      Parameters:
        mat     NEML small strain material model
        A       bar cross-sectional area
        l       bar length
      
      Optional:
        T       bar temperature as a function of time, defaults to 0 
    """
    self.mat = uniaxial.UniaxialModel(mat)
    self.A = A
    self.l = l
    self.T = T
    
    self.stress = np.zeros((1,))
    self.strain = np.zeros((1,))
    self.mstrain = np.zeros((1,))
    self.estrain = np.zeros((1,))
    self.tstrain = np.zeros((1,))
    self.energy = np.zeros((1,))
    self.dissipation = np.zeros((1,))
    self.temperature = np.array([self.T(0)])
    self.history = np.array([self.mat.init_store()])

    self.stress_next = 0.0
    self.strain_next = 0.0
    self.mstrain_next = 0.0
    self.estrain_next = 0.0
    self.tstrain_next = 0.0
    self.energy_next = 0.0
    self.dissipation_next = 0.0
    self.T_next = 0.0
    self.history_next = np.empty(self.history[0].shape)

  def update_force(self, d, t_np1, t_n):
    """
      Run a stress update, provide the force and the tangent

      Parameters:
        d       DELTA bar displacement
        t_np1   next time step
        t_n     last time step
    """
    self.T_next = self.T(t_np1)

    a_np1 = self.mat.alpha(self.T_next)
    a_n = self.mat.alpha(self.temperature[-1])
    dT = self.T_next - self.temperature[-1]

    self.tstrain_next = self.tstrain[-1] + dT * (a_np1 + a_n) / 2

    self.strain_next = d / self.l
    self.mstrain_next = self.strain_next - self.tstrain_next
    
    (self.stress_next, self.history_next, tangent, self.energy_next,
        self.dissipation_next) = self.mat.update(
            self.mstrain_next, self.mstrain[-1], 
            self.T_next, self.temperature[-1],
            t_np1, t_n,
            self.stress[-1], self.history[-1],
            self.energy[-1], self.dissipation[-1])

    self.estrain_next = self.mat.elastic_strains(self.stress_next, self.T_next,
        self.history_next)

    return self.stress_next * self.A, tangent * self.A / self.l

  def advance_step(self):
    """
      Actually append the speculative bar histories
    """
    self.stress = np.append(self.stress, self.stress_next)
    self.strain = np.append(self.strain, self.strain_next)
    self.mstrain = np.append(self.mstrain, self.mstrain_next)
    self.estrain = np.append(self.estrain, self.estrain_next)
    self.tstrain = np.append(self.tstrain, self.tstrain_next)
    self.energy = np.append(self.energy, self.energy_next)
    self.dissipation = np.append(self.dissipation, self.dissipation_next)
    self.temperature = np.append(self.temperature, self.T_next)
    self.history = np.vstack((self.history, self.history_next))
   
def newton(RJ, x0, verbose = False, rtol = 1.0e-6, atol = 1.0e-8, miter = 50,
    linesearch = 'none', bt_tau = 0.5, bt_c = 1.0e-4, fail_iter = True):
  """
    Manually-code newton-raphson so that I can output convergence info, if
    requested.

    Parameters:
      RJ        function return the residual + jacobian
      x0        initial guess

    Optional:
      verbose   verbose output
  """
  R, J = RJ(x0)
  nR = la.norm(R)
  nR0 = nR
  x = np.copy(x0)
  
  i = 0

  if verbose:
    print("Iter.\tnR\t\tnR/nR0\t\tcond\t\tlinesearch")
    print("%i\t%e\t%e\t" % (i, nR, nR / nR0))

  while (nR > rtol * nR0) and (nR > atol):
    a = la.solve(J, R)

    if linesearch == 'none':
      f = 1.0
    elif linesearch == 'backtracking':
      f = backtrack(RJ, R, J, x, -a, tau = bt_tau, c = bt_c, verbose = verbose)
    else:
      raise ValueError("Unknown linesearch type.")
    
    x -= (a * f)
    R, J = RJ(x)
    nR = la.norm(R)
    i += 1
    if verbose:
      print("%i\t%e\t%e\t%e\t%f" % (i, nR, nR / nR0,la.cond(J), f))
    if i > miter:
      if verbose:
        print("")
      if fail_iter:
        raise MaximumIterations()
      else:
        break
  
  if verbose:
    print("")

  return x

def backtrack(RJ, R0, J0, x, a, tau = 0.5, c = 1.0e-4, verbose = False):
  """
    Do a backtracking line search

    Parameters:
      RJ        residual/jacobian function
      R0        value of R, to save a feval
      J0        value of J, to save a feval
      x         point to start from
      a         direction

    Optional:
      tau       backtrack tau
      c         backtrack c
      verbose   verbose output
  """
  alpha = 1.0
  cv = la.norm(RJ(x + alpha * a)[0])
  while cv > la.norm(R0 + c * alpha * np.dot(J0, a)):
    alpha *= tau
    cv = la.norm(RJ(x + alpha * a)[0])

  return alpha



import numpy as np
import numpy.linalg as la
import scipy.optimize as opt
from neml.nlsolvers import newton

class UniaxialModel(object):
  """
    Takes a NEML model as input and performs the necessary calculations
    to give it a uniaxial response.
  """
  def __init__(self, nemlmodel, verbose = False, rtol = 1.0e-6, atol = 1.0e-10,
      miter = 20):
    """
      Parameters:
        nemlmodel:       the underlying 3D NEML model

      Keyword Args:
        verbose:         verbosity flag
        rtol:            newton relative tolerance
        atol:            newton absolute tolerance
        miter:           newton max iterations
    """
    self.model = nemlmodel
    self.verbose = verbose
    self.rtol = rtol
    self.atol = atol
    self.miter = miter

  @property
  def nstore():
    """
      Number of variables required to support the material update
    """
    return self.model.nstore + 5

  def init_store(self):
    """
      Initialize the variables required to support the material update
    """
    return np.concatenate((np.zeros((5,)), self.model.init_store()))

  @property
  def nhist():
    """
      True number of model history variables
    """
    return self.model.nhist

  def init_hist(self):
    """
      Initialize the model history
    """
    return self.model.init_hist()

  def alpha(self, T):
    """
      Returns the temperature-dependent thermal expansion coefficient

      Parameters:
        T       temperature
    """
    return self.model.alpha(T)

  def elastic_strains(self, s, T, h):
    """
      Returns the elastic strain for a given stress and temperature
    """
    s_vec = self.ss2sv(s)
    return self.model.elastic_strains(s_vec, T, h)[0]

  def update(self, e_np1, e_n, T_np1, T_n, t_np1, t_n, s_n, h_n, u_n, p_n):
    """
      Material update function
      
      Parameters:
        e_np1:       next strain
        e_n:         previous strain
        T_np1:       next temperature
        T_n:         previous temperature
        t_np1:       next time
        t_n:         previous time
        s_n:         previous stress
        h_n:         previous history
        u_n:         previous energy
        p_n:         previous dissipation
    """
    e_n_vec = self.es2ev(e_n, h_n)
    s_n_vec = self.ss2sv(s_n)
    h_n_vec = h_n[5:]
    
    def RJ(x):
      e_np1_vec = self.es2ev(e_np1, x)
      s_np1, h_np1, A_np1, u_np1, p_np1 = self.model.update_sd(
          e_np1_vec, e_n_vec, T_np1, T_n, t_np1, t_n, s_n_vec, h_n_vec,
          u_n, p_n)
      
      R = s_np1[1:]
      J = A_np1[1:,:][:,1:]

      return R, J

    x0 = h_n[:5]
    ef = newton(RJ, x0, verbose = self.verbose, rtol = self.rtol,
        atol = self.atol, miter = self.miter)
    
    e_np1_vec = self.es2ev(e_np1, ef)
    s_np1_vec, h_np1_vec, A_np1_vec, u_np1, p_np1 = self.model.update_sd(
        e_np1_vec, e_n_vec, T_np1, T_n, t_np1, t_n, s_n_vec, h_n_vec,
        u_n, p_n)
    h_np1 = np.concatenate((ef, h_np1_vec))
    s_np1 = s_np1_vec[0]

    # Faster than inverting
    sv = la.solve(A_np1_vec[1:,:][:,1:], A_np1_vec[1:,0])
    A_np1 = A_np1_vec[0,0] - np.dot(A_np1_vec[0,1:], sv)

    return s_np1, h_np1, A_np1, u_np1, p_np1

  def ss2sv(self, s):
    """
      Helper to convert scalar stress into stress vector

      Parameters:
        s:       scalar stress
    """
    vec = np.zeros((6,))
    vec[0] = s
    return vec

  def es2ev(self, e, h):
    """
      Helper to convert scalar strains into strain vectors

      Parameters:
        e:       scalar strain
        h:       history (stores other 5 strain components)
    """
    vec = np.zeros((6,))
    vec[0] = e
    vec[1:] = h[:5]
    return vec



#!/usr/bin/env python3

from neml import models, interpolate, elasticity, history
from neml.cp import crystallography, slipharden, sliprules, inelasticity, kinematics, singlecrystal
from neml.math import rotations, tensors

import unittest
import numpy as np
import numpy.linalg as la

import common

class CommonTangents(object):
  def test_no_rots(self):
    self.drive(self.model_no_rot)

  def drive(self, model):
    d_n = np.zeros((6,))
    w_n = np.zeros((3,))
    s_n = np.zeros((6,))
    h_n = model.init_store()
    t_n = 0.0
    u_n = 0.0
    p_n = 0.0

    for i in range(self.nsteps):
      t_np1 = t_n + self.dt
      d_np1 = d_n + self.Ddir * self.dt
      w_np1 = w_n + self.Wdir * self.dt

      s_np1, h_np1, A_np1, B_np1, u_np1, p_np1 = model.update_ld_inc(
          d_np1, d_n, w_np1, w_n, self.T, self.T, t_np1, t_n, s_n, h_n,
          u_n, p_n)

      A_num = common.differentiate(lambda d: model.update_ld_inc(d, d_n, w_np1, w_n, self.T,
        self.T, t_np1, t_n, s_n, h_n, u_n, p_n)[0], d_np1)

      # This is somewhat iffy
      print(A_np1)
      print(A_num)
      self.assertTrue(np.allclose(A_np1, A_num, rtol = 1.0e-3))

      B_num = common.differentiate(lambda w: model.update_ld_inc(d_np1, d_n, w, w_n, self.T,
        self.T, t_np1, t_n, s_n, h_n, u_n, p_n)[0], w_np1)

      # Again, why the loose tolerance?
      self.assertTrue(np.allclose(B_np1, B_num, rtol = 1.0e-3))

      s_n = np.copy(s_np1)
      h_n = np.copy(h_np1)
      d_n = np.copy(d_np1)
      w_n = np.copy(w_np1)
      t_n = t_np1
      u_n = u_np1
      p_n = p_np1

class CommonSolver(object):
  def test_jacobian(self):
    fn = lambda x: self.model.RJ(x, self.ts)[0]
    Jn = common.differentiate(fn, self.x)
    
    R, J = self.model.RJ(self.x, self.ts)

    print(J)
    print(Jn)

    print((J-Jn))
    
    self.assertTrue(np.allclose(J, Jn, rtol = 1.0e-4))

class TestSingleCrystal(unittest.TestCase, CommonTangents, CommonSolver):
  def setUp(self):
    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.kmodel = kinematics.StandardKinematicModel(self.emodel, self.imodel)

    self.model = singlecrystal.SingleCrystalModel(self.kmodel, self.L, 
        initial_rotation = self.Q)
    self.model_no_rot = singlecrystal.SingleCrystalModel(self.kmodel, self.L,
        initial_rotation = self.Q, update_rotation = False, verbose = False)

    self.T = 300.0
    self.stress_n = np.array([120.0,-60.0,170.0,35.0,80.0,-90.0])
    self.stress_np1 = np.array([15.0,-40.0,120.0,70.0,-10.0,-50.0])

    self.d = np.array([0.1,-0.2,0.25,0.11,-0.05,0.075])
    self.w = np.array([0.1,0.2,-0.2])

    self.strength_n = 25.0
    self.strength_np1 = 30.0

    self.S_np1 = tensors.Symmetric(common.usym(self.stress_np1))
    self.S_n = tensors.Symmetric(common.usym(self.stress_n))

    self.D = tensors.Symmetric(common.usym(self.d))
    self.W = tensors.Skew(common.uskew(self.w))

    self.strength_n = 25.0
    self.strength_np1 = 30.0

    self.H_n = history.History()

    self.H_n.add_scalar("strength")
    self.H_n.set_scalar("strength", self.strength_n)

    self.H_np1 = history.History()

    self.H_np1.add_scalar("strength")
    self.H_np1.set_scalar("strength", self.strength_np1)

    self.dt = 2.0

    self.fixed = self.kmodel.decouple(self.S_n, self.D, self.W, self.Q, 
        self.H_n, self.L, self.T, history.History())

    self.ts = singlecrystal.SCTrialState(self.D, self.W, self.S_n, self.H_n, self.Q, self.L, self.T, self.dt,
        self.fixed)

    self.x = np.zeros((self.model.nparams,))
    self.x[:6] = self.stress_np1
    self.x[6] = self.strength_np1

    self.Ddir = np.array([0.01,-0.005,-0.003,0.01,0.02,-0.003]) * 2
    self.Wdir = np.array([0.02,-0.03,0.01]) * 2

    self.nsteps = 10

  def test_strength(self):
    h = self.model.init_store()
    h[8] = self.strength_np1
    self.assertTrue(np.isclose(self.model.strength(h, 
      self.T), self.strength_np1 + self.tau0))

  def test_Fe(self):
    Qc = rotations.Orientation(12.0,35.0,61.0, angle_type = "degrees")

    h = self.model.init_store()
    h[:4] = Qc.quat
    h[4:8] = self.Q.quat
    h[8] = self.strength_np1

    Fe1 = self.model.Fe(self.stress_np1, h, self.T)

    Re = (Qc * self.Q.inverse()).to_matrix()
    E = common.usym(np.array(self.emodel.S_tensor(self.T, Qc
      ).dot(self.S_np1).data)) + np.eye(3)
    Fe2 = la.inv(np.dot(E, Re))

    self.assertTrue(np.allclose(Fe1,Fe2, rtol = 1e-3))

  def test_nhist(self):
    self.assertEqual(self.model.nstore, 9)

  def test_init_hist(self):
    h = self.model.init_store()
    
    self.assertTrue(np.allclose(h[:4], self.Q.quat))
    self.assertTrue(np.allclose(h[4:8], self.Q.quat))
    self.assertTrue(np.isclose(h[8],0.0))

  def test_residual(self):
    R, J = self.model.RJ(self.x, self.ts)

    Rtrue = np.zeros((self.model.nparams,))
    srate = self.kmodel.stress_rate(self.S_np1, self.D, self.W, self.Q, self.H_np1, self.L, self.T, self.fixed)

    hrate = np.array(self.kmodel.history_rate(self.S_np1, self.D, self.W, self.Q, self.H_np1, self.L, self.T, self.fixed))

    Rtrue[:6] = self.stress_np1 - self.stress_n - srate.data * self.dt
    Rtrue[6:] = np.array(self.H_np1) - np.array(self.H_n) - hrate * self.dt

    self.assertTrue(np.allclose(Rtrue, R))

  def test_set_active(self):
    q = rotations.Orientation(15.0,50.0,60.0, angle_type = "degrees")
    h = self.model.init_store()

    self.model.set_active_orientation(h, q)
    self.assertTrue(np.allclose(q.quat, 
      self.model.get_active_orientation(h).quat))

  def test_set_passive(self):
    q = rotations.Orientation(15.0,50.0,60.0, angle_type = "degrees")
    h = self.model.init_store()

    self.model.set_passive_orientation(h, q)
    self.assertTrue(np.allclose(q.quat, 
      self.model.get_passive_orientation(h).quat))

    self.assertTrue(np.allclose(q.quat, 
      self.model.get_active_orientation(h).inverse().quat))

class TestComplicatedCrystal(unittest.TestCase, CommonTangents, CommonSolver):
  def setUp(self):
    self.tau0_0 = 10.0
    self.tau_sat_0 = 50.0
    self.b_0 = 2.5

    self.tau0_1 = 15.0
    self.tau_sat_1 = 25.0
    self.b_1 = 1.0

    self.strengthmodel = slipharden.SumSlipSingleStrengthHardening(
        [
          slipharden.VoceSlipHardening(self.tau_sat_0, self.b_0, self.tau0_0),
          slipharden.VoceSlipHardening(self.tau_sat_1, self.b_1, self.tau0_1)
          ])

    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.kmodel = kinematics.StandardKinematicModel(self.emodel, self.imodel)

    self.model = singlecrystal.SingleCrystalModel(self.kmodel, self.L, 
        initial_rotation = self.Q)
    self.model_no_rot = singlecrystal.SingleCrystalModel(self.kmodel, self.L,
        initial_rotation = self.Q, update_rotation = False, verbose = False)

    self.T = 300.0
    self.stress_n = np.array([120.0,-60.0,170.0,35.0,80.0,-90.0])
    self.stress_np1 = np.array([15.0,-40.0,120.0,70.0,-10.0,-50.0])

    self.d = np.array([0.1,-0.2,0.25,0.11,-0.05,0.075])
    self.w = np.array([0.1,0.2,-0.2])

    self.strength_n = 25.0
    self.strength_np1 = 30.0

    self.S_np1 = tensors.Symmetric(common.usym(self.stress_np1))
    self.S_n = tensors.Symmetric(common.usym(self.stress_n))

    self.D = tensors.Symmetric(common.usym(self.d))
    self.W = tensors.Skew(common.uskew(self.w))

    self.strength_n_0 = 25.0
    self.strength_np1_0 = 30.0
    self.strength_n_1 = 15.0
    self.strength_np1_1 = 17.5

    self.H_n = history.History()
    self.H_n.add_scalar("strength0")
    self.H_n.set_scalar("strength0", self.strength_n_0)
    self.H_n.add_scalar("strength1")
    self.H_n.set_scalar("strength1", self.strength_n_1)

    self.H_np1 = history.History()
    self.H_np1.add_scalar("strength0")
    self.H_np1.set_scalar("strength0", self.strength_np1_0)
    self.H_np1.add_scalar("strength1")
    self.H_np1.set_scalar("strength1", self.strength_np1_1)

    self.dt = 2.0

    self.fixed = self.kmodel.decouple(self.S_n, self.D, self.W, self.Q, self.H_n, self.L, self.T,
        history.History())

    self.ts = singlecrystal.SCTrialState(self.D, self.W, self.S_n, self.H_n, self.Q, self.L, self.T, self.dt,
        self.fixed)

    self.x = np.zeros((self.model.nparams,))
    self.x[:6] = self.stress_np1
    self.x[6] = self.strength_np1

    self.Ddir = np.array([0.01,-0.005,-0.003,0.01,0.02,-0.003]) * 2
    self.Wdir = np.array([0.02,-0.03,0.01]) * 2

    self.nsteps = 10

  def test_nhist(self):
    self.assertEqual(self.model.nstore, 10)

class TestNyeStuffCrystal(unittest.TestCase):
  def setUp(self):
    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.k = 1.1

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b,
        self.tau0, k = self.k)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.kmodel = kinematics.StandardKinematicModel(self.emodel, self.imodel)

    self.model = singlecrystal.SingleCrystalModel(self.kmodel, self.L, 
        initial_rotation = self.Q)
    self.model_no_rot = singlecrystal.SingleCrystalModel(self.kmodel, self.L,
        initial_rotation = self.Q, update_rotation = False, verbose = False)

    self.T = 300.0
    self.stress_n = np.array([120.0,-60.0,170.0,35.0,80.0,-90.0])
    self.stress_np1 = np.array([15.0,-40.0,120.0,70.0,-10.0,-50.0])

    self.d = np.array([0.1,-0.2,0.25,0.11,-0.05,0.075])
    self.w = np.array([0.1,0.2,-0.2])

    self.strength_n = 25.0
    self.strength_np1 = 30.0

    self.S_np1 = tensors.Symmetric(common.usym(self.stress_np1))
    self.S_n = tensors.Symmetric(common.usym(self.stress_n))

    self.D = tensors.Symmetric(common.usym(self.d))
    self.W = tensors.Skew(common.uskew(self.w))

    self.strength_n = 25.0
    self.strength_np1 = 30.0

    self.H_n = history.History()

    self.H_n.add_scalar("strength")
    self.H_n.set_scalar("strength", self.strength_n)

    self.H_np1 = history.History()

    self.H_np1.add_scalar("strength")
    self.H_np1.set_scalar("strength", self.strength_np1)

    self.dt = 2.0

    self.fixed = self.kmodel.decouple(self.S_n, self.D, self.W, self.Q, 
        self.H_n, self.L, self.T, history.History())

    self.ts = singlecrystal.SCTrialState(self.D, self.W, self.S_n, self.H_n, self.Q, self.L, self.T, self.dt,
        self.fixed)

    self.x = np.zeros((self.model.nparams,))
    self.x[:6] = self.stress_np1
    self.x[6] = self.strength_np1

    self.Ddir = np.array([0.01,-0.005,-0.003,0.01,0.02,-0.003]) * 2
    self.Wdir = np.array([0.02,-0.03,0.01]) * 2

    self.nsteps = 10

    self.nye = np.array([[0.01,0.02,0.03],[1.2,1.1,1.3],[0.8,0.9,1.0]])

  def test_nhist(self):
    self.assertEqual(self.model.nstore, 18)

  def test_use_nye(self):
    self.assertTrue(self.model.use_nye)

  def test_set_nye(self):
    h = self.model.init_store()
    self.assertTrue(np.allclose(h[8:17], np.zeros((3,3)).flatten()))
    self.model.update_nye(h, self.nye)
    self.assertTrue(np.allclose(h[8:17], self.nye.flatten()))



import numpy as np
import numpy.linalg as la

import scipy.interpolate as inter
import scipy.optimize as opt
from numpy.polynomial.legendre import leggauss

import numpy.random as ra

from neml.nlsolvers import MaximumIterations, MaximumSubdivisions, newton, scalar_newton

class Driver(object):
  """
    Superclass of all drivers, basically just sets up history and reports
    results.
  """
  def __init__(self, model, verbose = False, rtol = 1.0e-6, atol = 1.0e-10,
      miter = 25, T_init = 0.0, no_thermal_strain = False):
    """
      Parameters:
        model:       material model to play with
        verbose:     verbose output
        rtol:        relative tolerance, where needed
        atol:        absolute tolerance, where needed
        miter:       maximum iterations, where needed
    """
    self.model = model
    self.verbose = verbose
    self.rtol = rtol
    self.atol = atol
    self.miter = miter
    self.nts = no_thermal_strain

    self.stress_int = [np.zeros((6,))]

    self.stored_int = [self.model.init_store()]
    self.T_int = [T_init]
    self.t_int = [0.0]

    self.u_int = [0.0]
    self.p_int = [0.0]

  @property
  def stress(self):
    return np.array(self.stress_int)

  @property
  def stored(self):
    return np.array(self.stored_int)

  @property
  def history(self):
    return self.stored[:,:self.model.nhist]

  @property
  def T(self):
    return np.array(self.T_int)

  @property
  def t(self):
    return np.array(self.t_int)

  @property
  def u(self):
    return np.array(self.u_int)

  @property
  def p(self):
    return np.array(self.p_int)


class Driver_sd(Driver):
  """
    Superclass of generic small strain drivers, contains generic step methods.
  """
  def __init__(self, *args, **kwargs):
    """
      Parameters:
        model:       material model to play with
        verbose:     verbose output
        rtol:        relative tolerance, where needed
        atol:        absolute tolerance, where needed
        miter:       maximum iterations, where needed
    """
    super(Driver_sd, self).__init__(*args, **kwargs)
    self.strain_int = [np.zeros((6,))]
    self.thermal_strain_int = [np.zeros((6,))]
    self.mechanical_strain_int = [np.zeros((6,))]

  def solve_try(self, RJ, x0, extra = []):
    """
      Try several different nonlinear solvers in the hope that at least
      one will converge

      Parameters:
        RJ:      function that returns the residual equations and associated
                 Jacobian
        x0:      initial guess
        extra:   list of extra solver functions of the type below
    """
    def s1(x0i):
      try:
        x = newton(RJ, x0i, verbose = self.verbose,
            rtol = self.rtol, atol = self.atol, miter = self.miter)
        return x, True
      except Exception:
        return np.zeros((12,)), False

    def s2(x0i):
      try:
        res = opt.root(RJ, x0i, jac = True, method = 'lm')
        return res.x, res.success
      except Exception:
        return np.zeros((12,)), False

    def s3(x0i):
      try:
        x = newton(RJ, x0i, verbose = self.verbose,
            rtol = self.rtol, atol = self.atol, miter = self.miter,
            linesearch = 'backtracking')
        return x, True
      except Exception:
        return np.zeros((12,)), False

    solvers = [s1,s3]
    guesses = [x0] + extra

    success = False
    for xi in guesses:
      for solv in solvers:
        x, success = solv(xi)
        if success:
          break
      if success:
        break
    
    if not success:
      raise MaximumIterations()

    return x

  @property
  def strain(self):
    return np.array(self.strain_int)

  @property
  def thermal_strain(self):
    return np.array(self.thermal_strain_int)

  @property
  def mechanical_strain(self):
    return np.array(self.mechanical_strain_int)

  def update_thermal_strain(self, T_np1):
    """
      Move the thermal strains to the next step

      Parameters:
        T_np1:      next temperature
    """
    if self.nts:
      return np.zeros((6,))
    else:
      dT = T_np1 - self.T_int[-1]
      a_np1 = self.model.alpha(T_np1)
      a_n = self.model.alpha(self.T_int[-1])

      return self.thermal_strain_int[-1] + dT * (a_np1 + a_n) / 2 * np.array([1.0,1,1,0,0,0])

  def strain_step(self, e_np1, t_np1, T_np1):
    """
      Take a strain-controlled step

      Parameters:
        e_np1:       next strain
        t_np1:       next time
        T_np1:       next temperature
    """
    enext = self.update_thermal_strain(T_np1)
    s_np1, h_np1, A_np1, u_np1, p_np1 = self.model.update_sd(e_np1 - enext,
        self.mechanical_strain_int[-1],
        T_np1, self.T_int[-1], t_np1, self.t_int[-1], self.stress_int[-1],
        self.stored_int[-1], self.u_int[-1], self.p_int[-1])

    self.strain_int.append(np.copy(e_np1))
    self.mechanical_strain_int.append(e_np1 - enext)
    self.thermal_strain_int.append(enext)
    self.stress_int.append(np.copy(s_np1))
    self.stored_int.append(np.copy(h_np1))
    self.T_int.append(T_np1)
    self.t_int.append(t_np1)
    self.u_int.append(u_np1)
    self.p_int.append(p_np1)

  def stress_step(self, s_np1, t_np1, T_np1):
    """
      Take a stress-controlled step

      Parameters:
        s_np1:       next stress
        t_np1:       next time
        T_np1:       next temperature
    """
    enext = self.update_thermal_strain(T_np1)
    def RJ(e):
      s, h, A, u, p = self.model.update_sd(e - enext, self.mechanical_strain_int[-1],
        T_np1, self.T_int[-1], t_np1, self.t_int[-1],
        self.stress_int[-1],
        self.stored_int[-1], self.u_int[-1], self.p_int[-1])
      R = s - s_np1
      return R, A

    if len(self.strain_int) > 1:
      inc = self.strain_int[-1] - self.strain_int[-2]
      extra = [self.strain_int[-1] + inc]
    else:
      extra = []

    e_np1 = self.solve_try(RJ, self.strain_int[-1], extra = extra)

    self.strain_step(e_np1, t_np1, T_np1)

  def erate_step(self, sdir, erate, t_np1, T_np1,
      einc_guess = None, ainc_guess = None):
    """
      Drive in a given stress direction at a prescribed strain rate, like
      an actual "stress controlled" experiment.

      Parameters:
        sdir:        stress direction
        erate:       strain rate (in the direction)
        t_np1:       next time
        T_np1:       next temperature
        einc_guess:  a guess at the strain increment
        ainc_guess:  a guess at the stress increment
    """
    sdir = sdir / la.norm(sdir)
    dt = t_np1 - self.t_int[-1]
    enext = self.update_thermal_strain(T_np1)

    def RJ(x):
      a = x[0]
      e_inc = x[1:]
      s, h, A, u, p = self.model.update_sd(self.strain_int[-1] + e_inc - enext,
          self.mechanical_strain_int[-1],
          T_np1, self.T_int[-1], t_np1, self.t_int[-1], self.stress_int[-1],
          self.stored_int[-1],
          self.u_int[-1], self.p_int[-1])

      R = np.zeros((7,))
      J = np.zeros((7,7))
      R[:6] = s - (sdir * a + self.stress_int[-1])
      R[6] = np.dot(e_inc, sdir) / dt - erate

      J[:6,0] = -sdir
      J[:6,1:] = A
      J[6,0] = 0.0
      J[6,1:] = sdir / dt

      return R, J

    x0 = np.zeros((7,))

    if einc_guess is not None:
      x0[1:] = einc_guess
    else:
      x0[1:] = sdir / 10000.0

    if ainc_guess is not None:
      x0[0] = ainc_guess
    else:
      x0[0] = 1.0

    x = self.solve_try(RJ, x0)
    e_np1 = self.strain_int[-1] + x[1:]

    self.strain_step(e_np1, t_np1, T_np1)

    return x[1:], x[0]

  def erate_einc_step(self, sdir, erate, einc, T_np1, **kwargs):
    """
      Similar to erate_step but specify the strain increment instead of the
      time increment.

      Parameters:
        sdir:        stress direction
        erate:       strain rate, in stress direction
        einc:        strain increment, in stress direction
        T_np1:       temperature at next time step
    """
    dt = einc / erate
    return self.erate_step(sdir, erate, self.t_int[-1] + dt, T_np1, **kwargs)

  def srate_sinc_step(self, sdir, srate, sinc, T_np1):
    """
      Similar to rate_step but specify the stress increment instead of the
      time increment.

      Parameters:
        sdir:        stress direction
        srate:       stress rate
        sinc:        stress increment
        T_np1:       temperature at next time step

    """
    if np.allclose(sdir, 0.0):
      s_np1 = self.stress_int[-1]
    else:
      s_np1 = self.stress_int[-1] + sdir / la.norm(sdir) * sinc
    if np.isclose(srate, 0.0):
      dt = 0.0
    else:
      dt = np.abs(np.dot(s_np1 - self.stress_int[-1], sdir) / srate)

    self.stress_step(s_np1, self.t_int[-1] + dt, T_np1)

  def strain_hold_step(self, i, t_np1, T_np1, q = 1.0, E = -1.0):
    """
      A special, mixed step which holds the strain in index i constant
      while holding the stress in the other directions to their previous
      values

      Parameters:
        i:           index to hold
        t_np1:       next time
        T_np1:       next temperature
        q:           follow up factor
        E:           Young's modulus to use -- must redo interface at some point
    """
    if not np.isclose(q, 1.0) and np.isclose(E, -1.0):
      raise ValueError("You must supply the Youngs modulus")

    enext = self.update_thermal_strain(T_np1)
    oset = sorted(list(set(range(6)) - set([i])))
    def RJ(e_np1):
      s, h, A, u, p = self.model.update_sd(e_np1 - enext,
          self.mechanical_strain_int[-1],
          T_np1, self.T_int[-1], t_np1, self.t_int[-1], self.stress_int[-1],
          self.stored_int[-1], self.u_int[-1], self.p_int[-1])

      R = np.zeros((6,))
      R[0] = (e_np1[i] - self.strain_int[-1][i]
          ) + (s[i] - self.stress_int[-1][i]) / E * (q - 1)
      R[1:] = s[oset] - self.stress_int[-1][oset]

      J = np.zeros((6,6))
      J[0,0] = 1.0
      J[0,:] += A[i,:] / E * (q - 1)
      J[1:,:] = A[oset,:][:]

      return R, J

    x0 = np.copy(self.strain_int[-1])

    e_np1 = self.solve_try(RJ, x0)

    self.strain_step(e_np1, t_np1, T_np1)

def uniaxial_test(model, erate, T = 300.0, emax = 0.05, nsteps = 250,
    sdir = np.array([1,0,0,0,0,0]), verbose = False,
    offset = 0.2/100.0, history = None, tdir = np.array([0,1,0,0,0,0]),
    rtol = 1e-6, atol = 1e-10, miter = 25):
  """
    Make a uniaxial stress/strain curve

    Parameters:
      model:            material model
      erate:            strain rate

    Keyword Args:
      T:                temperature, default 300.0
      emax:             maximum strain, default 5%
      nsteps:           number of steps to use, default 250
      sdir:             stress direction, default tension in x
      verbose:          whether to be verbose
      offset:           used to calculate yield stress
      history:          initial model history
      tdir:             transverse direction for Poisson's ratio

    Returns:
      dict:             results dictionary containing...


    **Results in dictionary:**
      ================= ============================================
      Name              Description
      ================= ============================================
      strain            strain in direction
      stress            stress in direction
      energy_density    strain energy density
      plastic_work      plastic dissipation
      youngs            young's modulus of initial curve
      yield             yield stress implied by curve
      poissons          poisson's ratio implied by non-axial strains
      ================= ============================================
  """
  e_inc = emax / nsteps
  driver = Driver_sd(model, verbose = verbose, T_init = T, rtol = rtol,
      atol = atol, miter = miter)
  if history is not None:
    driver.stored_int[0] = history

  strain = [0.0]
  stress = [0.0]
  for i in range(nsteps):
    if i == 0:
      einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T)
    else:
      einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
          einc_guess = einc, ainc_guess = ainc)
    strain.append(np.dot(driver.strain_int[-1], sdir))
    stress.append(np.dot(driver.stress_int[-1], sdir))

  strain = np.array(strain)
  stress = np.array(stress)

  # Calculate the yield stress and Young's modulus
  E = np.abs(stress[1]) / np.abs(strain[1])
  nu = -np.dot(driver.strain_int[1], tdir) / np.dot(
      driver.strain_int[1], sdir)
  sfn = inter.interp1d(np.abs(strain), np.abs(stress))
  tfn = lambda e: E * (e - offset)
  try:
    sYe = opt.brentq(lambda e: sfn(e) - tfn(e), 0.0, np.max(strain))
    sY = tfn(sYe)
  except Exception:
    sY = np.inf

  return {'strain': strain, 'stress': stress,
      'energy_density': np.copy(driver.u),
      'plastic_work': np.copy(driver.p),
      'youngs': E, 'yield': sY, 'poissons': nu,
      'history': driver.stored_int[-1]}

def strain_cyclic(model, emax, R, erate, ncycles, T = 300.0, nsteps = 50,
    sdir = np.array([1,0,0,0,0,0]), hold_time = None, n_hold = 25,
    verbose = False, check_dmg = False, dtol = 0.75):
  """
    Strain controlled cyclic test.

    Parameters:
      emax:         maximum strain
      R:            R = emin / emax
      erate:        strain rate to go at
      ncycles:      number of cycles
      T:            temperature, default 300

    Keyword Args:
      nsteps:       number of steps per half cycle
      sdir:         stress direction, defaults to x and tension first
      hold_time:    if None don't hold, if scalar then hold symmetrically top/bot
                    if an array specify different hold times for first direction
                    (default tension) and second direction
      n_hold:       number of steps to hold over
      verbose:      whether to be verbose
      check_dmg:    check to see if material damage exceeds dtol, stop the
                    simulation when that happens
      dtol:         damage to stop at

    Returns:
      dict:         results dictionary containing...

    **Results in dictionary:**
      ============= ========================
      Name          Description
      ============= ========================
      strain:       strain in direction
      stress:       stress in direction
      cycles:       list of cycle numbers
      max:          maximum stress per cycle
      min:          minimum stress per cycle
      mean:         mean stress per cycle
      ============= ========================
  """
  # Setup
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  emin = emax * R
  if hold_time:
    if np.isscalar(hold_time):
      hold_time = [hold_time, hold_time]
  else:
    hold_time = [0,0]

  # Setup results
  strain = [0.0]
  stress = [0.0]
  time = [0.0]
  cycles = []
  smax = []
  smin = []
  smean = []

  ecycle = []
  pcycle = []

  # First half cycle
  if verbose:
    print("Initial half cycle")
  e_inc = emax / nsteps
  try:
    for i in range(nsteps):
      if i == 0:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T)
      else:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T, einc_guess = einc,
            ainc_guess = ainc)
      if check_dmg:
        if driver.stored_int[-1][0] > dtol:
          raise Exception("Damage check exceeded")
      strain.append(np.dot(driver.strain_int[-1], sdir))
      stress.append(np.dot(driver.stress_int[-1], sdir))
      time.append(time[-1] + e_inc / erate)
  except Exception as e:
    print("Failed to make first half cycle")
    raise e

  # Begin cycling
  for s in range(ncycles):
    if verbose:
      print("Cycle %i" % s)

    try:
      # Tension hold
      if hold_time[0] > 0.0:
        dt = hold_time[0] / n_hold
        for i in range(n_hold):
          einc, ainc = driver.erate_step(sdir, 0.0, time[-1] + dt, T,
              einc_guess = np.zeros((6,)), ainc_guess = -1)
          if check_dmg:
            if driver.stored_int[-1][0] > dtol:
              raise Exception("Damage check exceeded")
          strain.append(np.dot(driver.strain_int[-1], sdir))
          stress.append(np.dot(driver.stress_int[-1], sdir))
          time.append(time[-1] + dt)

      si = len(driver.strain_int)
      e_inc = np.abs(emin - emax) / nsteps
      for i in range(nsteps):
        if i == 0:
          einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
              einc_guess = -einc, ainc_guess = -ainc)
        else:
          einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
              einc_guess = einc, ainc_guess = ainc)
        if check_dmg:
          if driver.stored_int[-1][0] > dtol:
            raise Exception("Damage check exceeded")
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))
        time.append(time[-1] + e_inc / erate)

      # Compression hold
      if hold_time[1] > 0.0:
        dt = hold_time[1] / n_hold
        for i in range(n_hold):
          einc, ainc = driver.erate_step(sdir, 0.0, time[-1] + dt, T,
              einc_guess = np.zeros((6,)), ainc_guess = -1)
          if check_dmg:
            if driver.stored_int[-1][0] > dtol:
              raise Exception("Damage check exceeded")
          strain.append(np.dot(driver.strain_int[-1], sdir))
          stress.append(np.dot(driver.stress_int[-1], sdir))
          time.append(time[-1] + dt)

      e_inc = np.abs(emax - emin) / nsteps
      for i in range(nsteps):
        if i == 0:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
              einc_guess = -einc, ainc_guess = -ainc)
        else:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
              einc_guess = einc, ainc_guess = ainc)
        if check_dmg:
          if driver.stored_int[-1][0] > dtol:
            raise Exception("Damage check exceeded")
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))
        time.append(time[-1] + e_inc / erate)

      # Calculate
      if np.isnan(max(stress[si:])) or np.isnan(min(stress[si:])):
        break

      cycles.append(s)
      smax.append(max(stress[si:]))
      smin.append(min(stress[si:]))
      smean.append((smax[-1]+smin[-1])/2)

      ecycle.append(driver.u_int[-1])
      pcycle.append(driver.p_int[-1])
    except Exception as e:
      break

  # Setup and return
  return {"strain": np.array(strain), "stress": np.array(stress),
      "cycles": np.array(cycles, dtype = int), "max": np.array(smax),
      "min": np.array(smin), "mean": np.array(smean),
      "energy_density": np.array(ecycle), "plastic_work": np.array(pcycle),
      "history": driver.stored_int[-1], "time": np.array(time)}

def strain_cyclic_extrapolated(model, emax, R, erate, ncycles, T = 300.0, nsteps = 50,
    sdir = np.array([1,0,0,0,0,0]), hold_time = None, n_hold = 25,
    verbose = False, check_dmg = False, dtol = 0.75, min_cycle=3, unit_extrapolate = 10,
    jump_delta_N=10, allowable_jump_stress=5.0):
  """
    Strain controlled cyclic test extrapolation.

    Extra Keyword Args:
      min_cycle                minimum cycles to start the extrapolation process
      unit_extrapolate         number of cycles to perform single cycle extrapolation
      jump_delta_N               number of cycles to jump
      allowable_jump_stress    extrapolate when stress jump is within this limit

    Returns:
      dict:         results dictionary containing...

    **Results in dictionary:**
      ============= ========================
      Name          Description
      ============= ========================
      cycles:       list of cycle numbers
      max:          maximum stress per cycle
      min:          minimum stress per cycle
      ============= ========================
  """
  # Setup
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  emin = emax * R
  if hold_time:
    if np.isscalar(hold_time):
      hold_time = [hold_time, hold_time]
  else:
    hold_time = [0,0]

  # Setup results
  strain = [0.0]
  stress = [0.0]
  time = [0.0]
  cycles = []
  smax = []
  smin = []
  smean = []

  ecycle = []
  pcycle = []

  # First half cycle
  if verbose:
    print("Initial half cycle")
  e_inc = emax / nsteps
  try:
    for i in range(nsteps):
      if i == 0:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T)
      else:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T, einc_guess = einc,
            ainc_guess = ainc)
      if check_dmg:
        if driver.stored_int[-1][0] > dtol:
          raise Exception("Damage check exceeded")

      strain.append(np.dot(driver.strain_int[-1], sdir))
      stress.append(np.dot(driver.stress_int[-1], sdir))
      time.append(time[-1] + e_inc / erate)

  except Exception as e:
    print("Failed to make first half cycle")
    raise e
  s = 0

# steps in one cycle
  if (hold_time[0] > 0) and (hold_time[1] == 0):
      steps = 2*nsteps + n_hold
  elif (hold_time[1] > 0) and (hold_time[0] == 0):
      steps = 2*nsteps + n_hold
  elif (hold_time[0] > 0) and (hold_time[1] > 0):
      steps = 2*nsteps + 2*n_hold
  else:
      steps = 2*nsteps

  extrapolate = False
  while s < ncycles:
    if verbose:
      print("Cycle %i" % s)

    if check_dmg:
      if driver.stored_int[-1][0] > dtol:
        print("Damage check exceeded")
        break

    if (s >= min_cycle) and (extrapolate == True):        # No extrapolation before min_cycle
        if (s <= unit_extrapolate):                       # single cycle jump for first unit_extrapolate cycles
            delta_N = 1
        else:
            delta_N = jump_delta_N                        # specified cycles to jump
        n = len(driver.stored_int)
        # extrapolating history
        pos_hist_last_last = driver.stored_int[n - 1 - steps]
        pos_hist_last      = driver.stored_int[n-1]
        dN_1 = cycles[-1] - cycles[-2]
        pos_extrapolated_history = pos_hist_last + (pos_hist_last - pos_hist_last_last)*delta_N/dN_1
        # extrapolating smax
        smax_last_last = smax[-2]
        smax_last = smax[-1]
        extrapolated_smax = smax_last + (smax_last - smax_last_last)*delta_N/dN_1
        # extrapolating smax
        smin_last_last = smin[-2]
        smin_last = smin[-1]
        extrapolated_smin = smin_last + (smin_last - smin_last_last)*delta_N/dN_1
        # criteria for extrapolation
        pos_stress_last_last = driver.stress_int[n - 1 - 2*steps]
        pos_stress_last      = driver.stress_int[n-1]
        pos_extrapolated_stress = pos_stress_last + (pos_stress_last - pos_stress_last_last)*delta_N/dN_1
        stress_jump = pos_extrapolated_stress[0] - pos_stress_last[0]
        if np.fabs(stress_jump) <= allowable_jump_stress:
            s = s + delta_N
            if s > ncycles:
                break
            driver.stored_int.append(pos_extrapolated_history)
            driver.stress_int.append(pos_extrapolated_stress)
            smax.append(extrapolated_smax)
            smin.append(extrapolated_smin)
            cycles.append(s)
            extrapolate = False
        else:
            extrapolate = False

    else:
        try:
          if hold_time[0] > 0.0:
            dt = hold_time[0] / n_hold
            for i in range(n_hold):
              einc, ainc = driver.erate_step(sdir, 0.0, time[-1] + dt, T,
                  einc_guess = np.zeros((6,)), ainc_guess = -1)
              if check_dmg:
                if driver.stored_int[-1][0] > dtol:
                  raise Exception("Damage check exceeded")
              strain.append(np.dot(driver.strain_int[-1], sdir))
              stress.append(np.dot(driver.stress_int[-1], sdir))
              time.append(time[-1] + dt)

          si = len(driver.strain_int)
          e_inc = np.abs(emin - emax) / nsteps
          for i in range(nsteps):
            if i == 0:
              einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
                  einc_guess = -einc, ainc_guess = -ainc)
            else:
              einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
                  einc_guess = einc, ainc_guess = ainc)
            if check_dmg:
              if driver.stored_int[-1][0] > dtol:
                raise Exception("Damage check exceeded")
            strain.append(np.dot(driver.strain_int[-1], sdir))
            stress.append(np.dot(driver.stress_int[-1], sdir))
            time.append(time[-1] + e_inc / erate)

          # Compression hold
          if hold_time[1] > 0.0:
            dt = hold_time[1] / n_hold
            for i in range(n_hold):
              einc, ainc = driver.erate_step(sdir, 0.0, time[-1] + dt, T,
                  einc_guess = np.zeros((6,)), ainc_guess = -1)
              if check_dmg:
                if driver.stored_int[-1][0] > dtol:
                  raise Exception("Damage check exceeded")
              strain.append(np.dot(driver.strain_int[-1], sdir))
              stress.append(np.dot(driver.stress_int[-1], sdir))
              time.append(time[-1] + dt)

          e_inc = np.abs(emax - emin) / nsteps
          for i in range(nsteps):
            if i == 0:
              einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
                  einc_guess = -einc, ainc_guess = -ainc)
            else:
              einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
                  einc_guess = einc, ainc_guess = ainc)
            if check_dmg:
              if driver.stored_int[-1][0] > dtol:
                raise Exception("Damage check exceeded")

            strain.append(np.dot(driver.strain_int[-1], sdir))
            stress.append(np.dot(driver.stress_int[-1], sdir))
            time.append(time[-1] + e_inc / erate)

          # Calculate
          if np.isnan(max(stress[si:])) or np.isnan(min(stress[si:])):
            break
          s += 1
          cycles.append(s)
          smax.append(max(stress[si:]))
          smin.append(min(stress[si:]))
          smean.append((smax[-1]+smin[-1])/2)
          ecycle.append(driver.u_int[-1])
          pcycle.append(driver.p_int[-1])
          extrapolate = True
        except Exception as e:
          break

   # Setup and return
  return {"cycles": np.array(cycles, dtype = int), "max": np.array(smax),
       "min": np.array(smin),"time": np.array(time)}


def strain_cyclic_followup(model, emax, R, erate, ncycles,
    q = 1.0, T = 300.0, nsteps = 50,
    sind = 0, hold_time = None, n_hold = 25,
    verbose = False, check_dmg = False, dtol = 0.75,
    logspace = False):
  """
    Strain controlled cyclic test with follow up.

    This is a "fallback" to the old version that does things by index
    so that I can use the index-based hold routine with follow up

    Parameters:
      emax:      maximum strain
      R:         R = emin / emax
      erate:     strain rate to go at
      ncycles:   number of cycles

    Keyword Args:
      q:         follow up factor
      T:         temperature, default 300
      nsteps:    number of steps per half cycle
      sind:      index to pull on
      hold_time: if None don't hold, if scalar then hold symmetrically top/bot
                 if an array specify different hold times for first direction
                 (default tension) and second direction
      n_hold:    number of steps to hold over
      verbose:   whether to be verbose
      check_dmg: check to see if damage exceeds a threshold
      dtol:      damage threshold
      logspace:  logspace the hold time steps (instead of linspace)

    Returns:
      dict:      dictionary of results...


    **Results in dictionary**
      ========= ========================
      Name      Description
      ========= ========================
      strain    strain in direction
      stress    stress in direction
      cycles    list of cycle numbers
      max       maximum stress per cycle
      min       minimum stress per cycle
      mean      mean stress per cycle
      ========= ========================
  """
  # Setup
  sdir = np.zeros((6,))
  sdir[sind] = 1.0

  res = uniaxial_test(model, erate, T = T, emax = 1.0e-4, nsteps = 2)
  E = res['youngs']

  driver = Driver_sd(model, verbose = verbose, T_init = T)
  emin = emax * R
  if hold_time:
    if np.isscalar(hold_time):
      hold_time = [hold_time, hold_time]
  else:
    hold_time = [0,0]

  # Setup results
  strain = [0.0]
  stress = [0.0]
  time = [0.0]
  cycles = []
  smax = []
  smin = []
  smean = []

  ecycle = []
  pcycle = []

  # First half cycle
  if verbose:
    print("Initial half cycle")
  e_inc = emax / nsteps
  try:
    for i in range(nsteps):
      if i == 0:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T)
      else:
        einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T, einc_guess = einc,
            ainc_guess = ainc)
      if check_dmg:
        if driver.stored_int[-1][0] > dtol:
          raise Exception("Damage check exceeded")
      strain.append(np.dot(driver.strain_int[-1], sdir))
      stress.append(np.dot(driver.stress_int[-1], sdir))
      time.append(time[-1] + e_inc / erate)
  except Exception as e:
    print("Failed to make first half cycle")
    raise e

  # Begin cycling
  for s in range(ncycles):
    if verbose:
      print("Cycle %i" % s)

    try:
      # Tension hold
      if hold_time[0] > 0.0:
        if logspace:
          dts = np.diff(np.logspace(0, np.log10(hold_time[0]), n_hold+1))
        else:
          dts = np.diff(np.linspace(0,hold_time[0],n_hold+1))
        #dt = hold_time[0] / n_hold
        for i, dt in enumerate(dts):
          driver.strain_hold_step(sind, time[-1] + dt, T,
              q = q, E = E)
          if check_dmg:
            if driver.stored_int[-1][0] > dtol:
              raise Exception("Damage check exceeded")
          strain.append(np.dot(driver.strain_int[-1], sdir))
          stress.append(np.dot(driver.stress_int[-1], sdir))
          time.append(time[-1] + dt)

      si = len(driver.strain_int)
      e_inc = np.abs(emin - emax) / nsteps
      for i in range(nsteps):
        if i == 0:
          einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
              einc_guess = np.zeros((6,)), ainc_guess = -1)
        else:
          einc, ainc = driver.erate_einc_step(-sdir, erate, e_inc, T,
              einc_guess = einc, ainc_guess = ainc)
        if check_dmg:
          if driver.stored_int[-1][0] > dtol:
            raise Exception("Damage check exceeded")
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))
        time.append(time[-1] + e_inc / erate)

      # Compression hold
      if hold_time[1] > 0.0:
        if logspace:
          dts = np.diff(np.logspace(0, np.log10(hold_time[1]), n_hold+1))
        else:
          dts = np.diff(np.linspace(0,hold_time[1],n_hold+1))
        for i, dt in enumerate(dts):
          driver.strain_hold_step(sind, time[-1] + dt, T,
              q = q, E = E)
          if check_dmg:
            if driver.stored_int[-1][0] > dtol:
              raise Exception("Damage check exceeded")
          strain.append(np.dot(driver.strain_int[-1], sdir))
          stress.append(np.dot(driver.stress_int[-1], sdir))
          time.append(time[-1] + dt)

      e_inc = np.abs(emax - emin) / nsteps
      for i in range(nsteps):
        if i == 0:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
              einc_guess = np.zeros((6,)), ainc_guess = 1.0)
        else:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
              einc_guess = einc, ainc_guess = ainc)
        if check_dmg:
          if driver.stored_int[-1][0] > dtol:
            raise Exception("Damage check exceeded")
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))
        time.append(time[-1] + e_inc / erate)

      # Calculate
      if np.isnan(max(stress[si:])) or np.isnan(min(stress[si:])):
        break

      cycles.append(s)
      smax.append(max(stress[si:]))
      smin.append(min(stress[si:]))
      smean.append((smax[-1]+smin[-1])/2)

      ecycle.append(driver.u_int[-1])
      pcycle.append(driver.p_int[-1])
    except Exception as e:
      break

  # Setup and return
  return {"strain": np.array(strain), "stress": np.array(stress),
      "cycles": np.array(cycles, dtype = int), "max": np.array(smax),
      "min": np.array(smin), "mean": np.array(smean),
      "energy_density": np.array(ecycle), "plastic_work": np.array(pcycle),
      "history": driver.stored_int[-1], "time": np.array(time)}


def stress_cyclic(model, smax, R, srate, ncycles, T = 300.0, nsteps = 50,
    sdir = np.array([1,0,0,0,0,0]), hold_time = None, n_hold = 10,
    verbose = False, etol = 0.1):
  """
    Stress controlled cyclic test.

    Parameters:
      emax:         maximum stress
      R:            R = smin / smax
      erate:        strain rate to go at
      ncycles:      number of cycles

    Keyword Args:
      T:            temperature, default 300
      nsteps:       number of steps per half cycle
      sdir:         stress direction, defaults to x and tension first
      hold_time:    if None don't hold, if scalar then hold symmetrically top/bot
                    if an array specify different hold times for first direction
                    (default tension) and second direction
      n_hold:       number of steps to hold over
      verbose:      whether to be verbose
      etol:         stop if the strain increment exceeds this threshold

    Returns:
      dict:         dictionary of results


    **Results in dictionary:**
      ============= ========================
      Name          Description
      ============= ========================
      strain        strain in direction
      stress        stress in direction
      cycles        list of cycle numbers
      max           maximum strain per cycle
      min           minimum strain per cycle
      mean          mean strain per cycle
      ============= ========================
  """
  # Setup
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  smin = smax * R
  if hold_time:
    if np.isscalar(hold_time):
      hold_time = [hold_time, hold_time]

  # Setup results
  strain = [0.0]
  stress = [0.0]
  cycles = []
  emax = []
  emin = []
  emean = []

  ecycle = []
  pcycle = []

  # First half cycle
  s_inc = smax / nsteps
  if verbose:
    print("First half cycle")
  for i in range(nsteps):
    driver.srate_sinc_step(sdir, srate, s_inc, T)
    strain.append(np.dot(driver.strain_int[-1], sdir))
    stress.append(np.dot(driver.stress_int[-1], sdir))

  # Begin cycling
  for s in range(ncycles):
    quit = False
    if verbose:
      print("Cycle %i" % s)
    si = len(driver.strain_int)
    # Hold, if requested
    if hold_time and (hold_time[0] > 0.0):
      ht = hold_time[0]
      dt = ht / n_hold
      for i in range(n_hold):
        try:
          driver.stress_step(driver.stress_int[-1], driver.t_int[-1] + dt, T)
        except:
          quit = True
          break
        if la.norm(driver.strain_int[-1] - driver.strain_int[-2]) > etol:
          quit = True
          break
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))

    if quit:
      break

    s_inc = (smin - smax) / nsteps
    for i in range(nsteps):
      try:
        driver.srate_sinc_step(sdir, srate, s_inc, T)
      except:
        quit = True
        break
      if la.norm(driver.strain_int[-1] - driver.strain_int[-2]) > etol:
        quit = True
        break
      strain.append(np.dot(driver.strain_int[-1], sdir))
      stress.append(np.dot(driver.stress_int[-1], sdir))

    if quit:
      break

    # Hold, if requested
    if hold_time and (hold_time[1] > 0.0):
      ht = hold_time[1]
      dt = ht / n_hold
      for i in range(n_hold):
        try:
          driver.stress_step(driver.stress_int[-1], driver.t_int[-1] + dt, T)
        except:
          quit = True
          break
        if la.norm(driver.strain_int[-1] - driver.strain_int[-2]) > etol:
          quit = True
          break
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))

    if quit:
      break

    s_inc = (smax - smin) / nsteps
    for i in range(nsteps):
      try:
        driver.srate_sinc_step(sdir, srate, s_inc, T)
      except:
        quit = True
        break
      if la.norm(driver.strain_int[-1] - driver.strain_int[-2]) > etol:
        quit = True
        break
      strain.append(np.dot(driver.strain_int[-1], sdir))
      stress.append(np.dot(driver.stress_int[-1], sdir))

    if quit:
      break

    # Calculate
    cycles.append(s)
    emax.append(max(strain[si:]))
    emin.append(min(strain[si:]))
    emean.append((emax[-1]+emin[-1])/2)
    ecycle.append(driver.u_int[-1])
    pcycle.append(driver.p_int[-1])

  # Setup and return
  return {"strain": np.array(strain), "stress": np.array(stress),
      "cycles": np.array(cycles, dtype = int), "max": np.array(emax),
      "min": np.array(emin), "mean": np.array(emean),
      "energy_density": np.array(ecycle), "plastic_work": np.array(pcycle),
      "time": np.array(driver.t_int)}

def stress_relaxation(model, emax, erate, hold, T = 300.0, nsteps = 250,
    nsteps_up = 50, index = 0, tc = 1.0,
    verbose = False, logspace = False, q = 1.0):
  """
    Simulate a stress relaxation test.

    Parameters:
      model:         material model
      emax :         maximum strain to attain
      erate:         strain rate to take getting there
      hold:          hold time

    Keyword Args:
      T:             temperature
      nsteps:        number of steps to relax over
      nsteps_up:     number of steps to take getting up to stress
      index:         direction to pull in, default x tension
      tc:            1.0 for tension -1.0 for compression
      verbose:       whether to be verbose
      logspace:      log space the relaxation timesteps
      q:             follow up factor

    Results:
      dict:          dictionary of results

    **Results in dictionary:**
      ============== ======================
      Name           Description
      ============== ======================
      time           time
      strain         strain
      stress         stress
      rtime          relaxation time
      rrate          stress relaxation rate
      ============== ======================
  """
  # Setup
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  time = [0]
  strain = [0]
  stress = [0]

  res = uniaxial_test(model, erate, T = T, emax = 1.0e-4, nsteps = 2)
  E = res['youngs']

  # Ramp up
  if verbose:
    print("Ramp up")
  sdir = np.zeros((6,))
  sdir[index] = tc
  einc = emax / nsteps_up
  for i in range(nsteps_up):
    if i == 0:
      eincg, ainc = driver.erate_einc_step(sdir, erate, einc, T)
    else:
      eincg, ainc = driver.erate_einc_step(sdir, erate, einc, T,
          einc_guess = eincg, ainc_guess = ainc)
    time.append(driver.t[-1])
    strain.append(np.dot(driver.strain_int[-1],sdir))
    stress.append(np.dot(driver.stress_int[-1],sdir))

  ri = len(driver.strain_int)

  if verbose:
    print("Hold")
  if logspace:
    ts = np.logspace(0, np.log10(hold), num = nsteps+1)
    dts = np.diff(ts)
  else:
    dt = hold / nsteps
    dts = [dt] * nsteps
  for i, dt in enumerate(dts):
    driver.strain_hold_step(index, driver.t_int[-1] + dt, T,
        q = q, E = E)
    time.append(driver.t_int[-1])
    strain.append(np.dot(driver.strain_int[-1],sdir))
    stress.append(np.dot(driver.stress_int[-1],sdir))

  time = np.array(time)
  strain = np.array(strain)
  stress = np.array(stress)
  rrate = -np.diff(stress[ri:]) / np.diff(time[ri:])

  return {'time': np.copy(time), 'strain': np.copy(strain),
      'stress': np.copy(stress), 'rtime': np.copy(time[ri:-1] - time[ri]),
      'rrate': np.copy(rrate), 'rstress': np.copy(stress[ri:-1]),
      'rstrain': np.copy(strain[ri:-1])}

def creep(model, smax, srate, hold, T = 300.0, nsteps = 250,
    nsteps_up = 150, sdir = np.array([1,0,0,0,0,0]), verbose = False,
    logspace = False, history = None, elimit = 1.0, check_dmg = False,
    dtol = 0.75):
  """
    Simulate a creep test

    Parameters:
      model:         material model
      smax:          stress to attain
      srate:         stress rate to take getting there
      hold:          total hold time

    Keyword Args:
      T:             temperature
      nsteps:        number of steps over relaxation period
      nsteps_up:     number of steps to get to stress value
      sdir:          stress direction, defaults to x-tension
      verbose:       whether to be verbose
      logspace:      if true logspace the time steps
      history:       use damaged material
      check_dmg:     check damage as a break condition
      dtol:          damage to define failure at

    Returns:
      dict:          results dictionary
  """
  # Setup
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  if history is not None:
    driver.stored_int[0] = history
  time = [0]
  strain = [0]
  stress = [0]

  # Ramp up
  sinc = float(smax) / nsteps_up
  for i in range(nsteps_up):
    driver.srate_sinc_step(sdir, srate, sinc, T)
    time.append(driver.t[-1])
    strain.append(np.dot(driver.strain_int[-1],sdir))
    stress.append(np.dot(driver.stress_int[-1],sdir))

  ri = len(driver.strain_int)

  t0 = time[-1]
  if logspace:
    ts = np.logspace(0, np.log10(hold), num = nsteps) + t0
  else:
    ts = np.linspace(0,hold, num = nsteps) + t0

  failed = False
  for t in ts:
    # You can exceed the creep life of the sample doing this...
    # Need to allow a non-convergent result to break
    try:
      driver.stress_step(driver.stress_int[-1], t, T)
    except:
      failed = True
      break
    if np.any(np.isnan(driver.strain_int[-1])):
      failed = True
      break
    if np.any(np.abs(driver.strain_int[-1]) > elimit):
      failed = True
      break

    ed = np.dot(driver.strain_int[-1],sdir)
    if ed < strain[-1]:
      failed = True
      break

    if check_dmg:
      if driver.stored_int[-1][0] > dtol:
        failed = True
        break

    time.append(t)
    strain.append(ed)
    stress.append(np.dot(driver.stress_int[-1],sdir))

  time = np.array(time)
  strain = np.array(strain)
  stress = np.array(stress)
  rrate = np.diff(strain[ri:]) / np.diff(time[ri:])
  if len(strain) > ri +1:
    rstrain = strain[ri:] - strain[ri]
    rtime = time[ri:] - time[ri]
  else:
    rstrain = []
    rtime = []

  return {'time': np.copy(time), 'strain': np.copy(strain),
      'stress': np.copy(stress), 'rtime': np.copy(rtime[:-1]),
      'rrate': np.copy(rrate), 'rstrain': np.copy(rstrain[:-1]),
      'tstrain': np.copy(strain[ri:-1]),
      'history': np.array(driver.stored_int), 'failed': failed}

def thermomechanical_strain_raw(model, time, temperature, strain,
    sdir = np.array([1,0,0,0,0,0.0]), verbose = False, substep = 1):
  """
    Directly drive a model using the output of a strain controlled
    thermomechanical test

    Parameters:
      model:         material model
      time:          list of times
      temperature:   list of temperatures
      strain:        list of strains

    Keyword Args:
      sdir:          direction of stress
      verbose:       print information
      substep:       take substep steps per data point

    Returns:
      dict:          results dictionary
  """
  stress = np.zeros((len(time),))
  mechstrain = np.zeros((len(time),))
  driver = Driver_sd(model, verbose = verbose, T_init = temperature[0])

  einc = None
  ainc = None

  for i in range(1,len(stress)):
    quit = False
    for k in range(substep):
      ei_np1 = (strain[i] - strain[i-1]) / substep * (k+1) + strain[i-1]
      ti_np1 = (time[i] - time[i-1]) / substep * (k+1) + time[i-1]
      Ti_np1 = (temperature[i] - temperature[i-1]) / substep * (k+1) + temperature[i-1]

      ei_n = (strain[i] - strain[i-1]) / substep * (k) + strain[i-1]
      ti_n = (time[i] - time[i-1]) / substep * (k) + time[i-1]
      Ti_n = (temperature[i] - temperature[i-1]) / substep * (k) + temperature[i-1]

      erate = (ei_np1 - ei_n) / (ti_np1 - ti_n)
      try:
        if i == 1:
          einc, ainc = driver.erate_step(sdir, erate, ti_np1, Ti_np1)
        else:
          einc, ainc = driver.erate_step(sdir, erate, ti_np1, Ti_np1,
              einc_guess = einc, ainc_guess = ainc)
      except MaximumIterations:
        quit = True
        break
    if quit:
      break

    stress[i] = np.dot(driver.stress_int[-1], sdir)
    mechstrain[i] = np.dot(driver.thermal_strain_int[-1], sdir)

  return {'time': np.copy(time)[:i], 'temperature': np.copy(temperature)[:i], 'strain': np.copy(strain)[:i],
      'stress': np.copy(stress)[:i], 'mechanical strain': np.copy(mechstrain)[:i]}

def rate_jump_test(model, erates, T = 300.0, e_per = 0.01, nsteps_per = 100,
    sdir = np.array([1,0,0,0,0,0]), verbose = False, history = None,
    strains = None):
  """
    Model a uniaxial strain rate jump test

    Parameters:
      model:         material model
      erate:         list of strain rates

    Keyword Args:
      T:             temperature, default 300.0
      e_per:         how much straining to do for each rate
      nsteps_per:    number of steps per strain rate
      sdir:          stress direction, default tension in x
      verbose:       whether to be verbose
      history:       prior model history
      strains:       manual definition of jumps

    Returns:
      dict:          results dictionary
  """
  e_inc = e_per / nsteps_per
  driver = Driver_sd(model, verbose = verbose, T_init = T)
  if history is not None:
    driver.stored_int[0] = history
  strain = [0.0]
  stress = [0.0]

  if strains is None:
    for erate in erates:
      for i in range(nsteps_per):
        if i == 0:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T)
        else:
          einc, ainc = driver.erate_einc_step(sdir, erate, e_inc, T,
              einc_guess = einc, ainc_guess = ainc)
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))
  else:
    incs = np.diff(np.insert(strains,0,0)) / nsteps_per
    for e,erate,inc in zip(strains,erates,incs):
      while strain[-1] < e:
        einc, ainc = driver.erate_einc_step(sdir, erate, inc, T)
        strain.append(np.dot(driver.strain_int[-1], sdir))
        stress.append(np.dot(driver.stress_int[-1], sdir))

  strain = np.array(strain)
  stress = np.array(stress)

  return {'strain': strain, 'stress': stress,
      'energy_density': np.copy(driver.u),
      'plastic_work': np.copy(driver.p)}


def isochronous_curve(model, time, T = 300.0, emax = 0.05, srate = 1.0,
    ds = 10.0, max_cut = 4, nsteps = 250, history = None,
    check_dmg = False, dtol = 0.75):
  """
    Generates an isochronous stress-strain curve at the given time and
    temperature.

    Parameters:
      model:     material model
      time:      relevant time

    Keyword Args:
      T:         temperature
      emax:      maximum strain to attain on the curves
      srate:     stress rate to use in the ramp part
      ds:        stress increment along the curve
      max_cut:   adaptive refinement
      nsteps:    number of creep steps
      history:   start with a non-zero initial history
      check_dmg: stop if damage exceeds a threshold
      dtol:      damage threshold
  """
  def strain(stress):
    res = creep(model, stress, srate, time, T = T, nsteps = nsteps,
        history = history, check_dmg = check_dmg, dtol = dtol)
    return res['tstrain'][-1], res['failed']

  strains = [0.0]
  stresses = [0.0]
  ncut = 0
  failed = False
  try:
    while strains[-1] < emax:
      target = stresses[-1] + ds
      try:
        enext, failed = strain(target)
        if failed:
          break
        stresses.append(target)
        strains.append(enext)
      except Exception:
        ncut += 1
        if ncut > max_cut:
          raise MaximumSubdivisions()
        ds /= 2
  except MaximumSubdivisions:
    # We were quite aggressive, so assume the curve goes flat
    stresses.append(stresses[-1])
    strains.append(emax)

  if failed:
    stresses.append(stresses[-1])
    strains.append(emax)

  # Now interpolate back the last strain point
  iff = inter.interp1d(strains, stresses)
  ls = iff(emax)
  strains[-1] = emax
  stresses[-1] = ls

  return {'strain': np.copy(strains), 'stress': np.copy(stresses)}

def offset_stress(e, s, eo = 0.2/100.0):
  """
    Helper function to generate yield stress from offset stress/strain data

    Parameters:
      e:     strain data
      s:     stress data
      eo:    strain offset
  """
  iff = inter.interp1d(e, s)
  E = s[1] / e[1]

  eoff = opt.brentq(lambda e: iff(e) - E * (e - eo), 0.0,np.max(e))
  soff = iff(eoff)

  return soff

def classify(ua, ub, pa, pb, e1a, e1b, e2a, e2b, rtol = 1.0e-4, atol = 1.0e-10):
  """
    Classify a model as elastic, elastic shakedown, plastic shakedown,
    or ratcheting.

    Parameters:
      ua:    cycle a internal energy
      ub:    cycle b internal energy
      pa:    cycle a plastic dissipation
      pb:    cycle b plastic dissipation
      ea:    cycle a strain
      eb:    cycle b strain
      rtol:  relative tolerance
      atol:  absolute tolerance
  """
  if np.abs(pb) < atol:
    return 'elastic'
  elif np.abs(ub-ua) < rtol * np.abs(ub):
    return 'elastic shakedown'
  elif (np.abs(e1b-e1a) < rtol * np.abs(e1b)) and (np.abs(e2b-e2a) < rtol * np.abs(e2b)):
    return 'plastic shakedown'
  else:
    return 'ratcheting'

def def_grad_driver(model, F, tmax, nsteps, T = 300.0):
  """
    Basic large deformation driver pushing a model through the deformation
    gradient as a function of time

    Parameters:
      model     model to use
      F         deformation gradient as a function of time
      tmax      maximum time
      nsteps    number of load steps to take

    Optional:
      T         temperature
  """
  time = [0.0]
  stress = [np.zeros((6,))]
  deform = [F(0.0)]
  D_hist = [np.zeros((6,))]
  W_hist = [np.zeros((3,))]
  history = [model.init_store()]
  energy = [0.0]
  dissipation = [0.0]

  t = 0.0
  dt = tmax / nsteps

  for i in range(nsteps):
    t += dt
    F_np1 = F(t)
    L = np.dot((F_np1 - deform[-1]), la.inv(F_np1))
    dD = sym(0.5*(L + L.T))
    dW = skew(0.5*(L - L.T))

    s_np1, h_np1, A_np1, B_np1, u_np1, p_np1 = model.update_ld_inc(
        D_hist[-1] + dD, D_hist[-1],
        W_hist[-1] + dW, W_hist[-1],
        T, T, t, time[-1], stress[-1],
        history[-1], energy[-1], dissipation[-1])

    time.append(t)
    stress.append(s_np1)
    deform.append(F_np1)
    D_hist.append(D_hist[-1] + dD)
    W_hist.append(W_hist[-1] + dW)
    history.append(h_np1)
    energy.append(u_np1)
    dissipation.append(p_np1)

  return {'time': time, 'stress': stress, 'F': deform, 'D': D_hist,
      'W': W_hist, 'history': history, 'energy': energy,
      'dissipation': dissipation}

def sym(A):
  """
    Take a symmetric matrix to the Mandel convention vector.
  """
  return np.array([A[0,0], A[1,1], A[2,2], np.sqrt(2)*A[1,2],
    np.sqrt(2)*A[0,2], np.sqrt(2)*A[0,1]])

def skew(A):
  """
    Take a skew matrix to my vector convention.
  """
  return np.array([-A[1,2], A[0,2], -A[0,1]])
